"""Tools for profiling code."""

import cProfile
from functools import wraps
import io
import pstats
import typing as t


AnyFn = t.Any


def profile(func: AnyFn) -> AnyFn:
    """Decorator that profiles function."""
    @wraps(func)
    def wrapper(*args: t.Any, **kwargs: t.Any) -> t.Any:
        profile = cProfile.Profile()
        profile.enable()

        result = func(*args, **kwargs)

        profile.disable()
        stream = io.StringIO()
        cumulative = pstats.SortKey.CUMULATIVE
        stats = pstats.Stats(profile, stream=stream).sort_stats(cumulative)
        stats.print_stats()
        print(stream.getvalue())
        return result
    return wrapper



# pylint: disable=missing-function-docstring
"""Test generator.py."""

from pathlib import Path

from slipbox.generator import output_directory_proxy as proxy


def test_proxy_rolls_back_changes_on_error(tmp_path: Path) -> None:
    present = tmp_path/"present.txt"
    present.write_text("present")

    try:
        with proxy(tmp_path):
            assert False
    except AssertionError:
        pass

    assert present in tmp_path.iterdir()
    assert present.read_text() == "present"


def test_proxy_only_applies_changes_after_block(tmp_path: Path) -> None:
    with proxy(tmp_path) as path:
        (path/"hello.txt").write_text("hello")
        assert not list(tmp_path.iterdir())

    hello = tmp_path/"hello.txt"
    assert hello in tmp_path.iterdir()
    assert hello.read_text() == "hello"


def test_proxy_clears_previous_contents(tmp_path: Path) -> None:
    (tmp_path/"foo").touch()
    (tmp_path/"bar").mkdir()

    with proxy(tmp_path) as path:
        (path/"present.txt").write_text("present")

    present = tmp_path/"present.txt"
    assert list(tmp_path.iterdir()) == [present]
    assert present.read_text() == "present"



"""slipbox CLI commands."""

import configparser
from pathlib import Path
import sys

from . import check
from .app import App, error, require_init, RootlessApp
from .config import Config
from .processor import METADATA_TEMPLATES


@require_init
def show_info(app: App) -> None:
    """Show note info."""
    note_id = int(app.args["note_id"])
    sql = "SELECT title, filename FROM Notes WHERE id = ?"
    cur = app.database.cursor()
    cur.execute(sql, (note_id,))
    note = cur.fetchone()

    if note is None:
        error(f"note {note_id} does not exist")

    print(note_id)
    print(note[0])
    print(note[1])


@require_init
def check_notes(app: App) -> None:
    """Check for isolated notes and invalid links."""
    if not check.check_notes(app):
        sys.exit(65)


def list_supported_formats(_: RootlessApp) -> None:
    """List supported input file formats."""
    for key in METADATA_TEMPLATES:
        print(f"*{key}")


def init(app: RootlessApp) -> None:
    """Initialize notes directory."""
    if app.root is not None:
        root = str(app.root.resolve())
        error(f"slipbox has already been initialized in {root}")

    config = app.args.get("config")
    if config is not None:
        try:
            app.config = Config.from_file(Path(config))
        except configparser.Error:
            error(f"invalid config file: {config}")

    app.root = Path().resolve()
    hidden = app.root/".slipbox"
    hidden.mkdir(parents=True, exist_ok=True)
    app.config.write(hidden/"config.cfg")

    if not app.args.get("quiet"):
        print(f"slipbox initialized in {app.root.resolve()!s}")
        print("You can configure slipbox by editing .slipbox/config.cfg.")



"""Test page.py."""

import sqlite3
import typing as t

import pytest
from slipbox import page
from slipbox.app import App, startup
from slipbox.database import migrate
from slipbox.dependencies import check_requirements

SQL = """
    PRAGMA foreign_keys=ON;
    INSERT INTO Files (filename) VALUES ('test.md');
    INSERT INTO Notes (id, title, filename, html) VALUES
        (0, '0', 'test.md', '<section><h1>0</h1><p>Foo.</p></section>'),
        (1, '1', 'test.md', '<section><h1>1</h1><p>Bar.</p></section>'),
        (2, '2', 'test.md', '<section><h1>2</h1><p>Baz.</p></section>');
    INSERT INTO Tags (tag, id) VALUES
        ('#test', 0),
        ('#test', 1),
        ('##test', 2);
    INSERT INTO Links (src, dest) VALUES
        (0, 0),
        (1, 1),
        (2, 2);
    INSERT INTO Bibliography (key, html) VALUES
        ('ref-test', 'Reference text.');
    INSERT INTO Citations (note, reference) VALUES
        (0, 'ref-test');
"""


@pytest.fixture
def mock_db() -> t.Iterable[sqlite3.Connection]:
    """Create an empty mock database with all the necessary tables."""
    with sqlite3.connect(":memory:") as con:
        migrate(con)
        yield con


def test_render_references(mock_db: sqlite3.Connection) -> None:
    """Check render_references output."""
    conn = mock_db
    conn.executescript(SQL)
    html = page.render_references(conn)
    # pylint: disable=line-too-long
    assert html == """<section id="references" class="level1" title="References" role="doc-bibliography">
<h1>References</h1>
<dl>
<dt role="doc-biblioentry"><a href="#ref-test">[@test]</a></dt>
<dd>Reference text.</dd>
</dl>
</section>
"""  # noqa


def test_render_tags(mock_db: sqlite3.Connection) -> None:
    """Check render_tags output."""
    conn = mock_db
    conn.executescript(SQL)
    html = page.render_tags(conn)
    assert html == """<section id="tags" title="Tags" class="level1">
<h1>Tags</h1>
<ul>
<li><a href="#tags/#test">##test</a> (1)</li>
<li><a href="#tags/test">#test</a> (2)</li>
</ul>

</section>
"""


def test_render_tag_page(mock_db: sqlite3.Connection) -> None:
    """Check render_tag_page output."""
    conn = mock_db
    conn.executescript(SQL)
    html = page.render_tag_page(conn, "#test")
    assert html == """<section id='tags/test' title='#test' class='level1'>
  <h1>
    #test
  </h1>
  <ul class="slipbox-list">
    <li value="0">0 – <a href="#0">0</a></li>
    <li value="1">1 – <a href="#1">1</a></li>
  </ul>
</section>"""


def test_render_reference_page(mock_db: sqlite3.Connection) -> None:
    """Check render_reference_page output."""
    conn = mock_db
    conn.executescript(SQL)
    html = page.render_reference_page(conn, "ref-test")
    assert html == """<section id='ref-test' title='ref-test' class='level1'>
  <h1>
    @test
  </h1>
  <p>
    Reference text.
  </p>
  <ul class="slipbox-list">
    <li value="0">0 – <a href="#0">0</a></li>
  </ul>
</section>"""


@pytest.mark.skipif(
    not check_requirements(startup({})),
    reason="missing requirements",
)
def test_generate_index(mock_db: sqlite3.Connection, app: App) -> None:
    """Sanity check."""
    app.database = mock_db
    page.generate_index(app, app.root)



"""Test dependencies.py."""

from slipbox.app import RootlessApp
from slipbox.dependencies import has_dot, has_pandoc


def test_has_dot_not_installed(app_without_root: RootlessApp) -> None:
    """has_dot should return False if dot is not installed."""
    app = app_without_root
    app.config.dot = ""
    assert not has_dot(app)


def test_has_pandoc_not_installed(app_without_root: RootlessApp) -> None:
    """has_pandoc should return False if pandoc is not installed."""
    app = app_without_root
    app.config.pandoc = ""
    assert not has_pandoc(app)



"""Testing fixtures."""

from pathlib import Path
import typing as t

import pytest

from slipbox.app import App, RootlessApp, startup
from slipbox.commands import init


@pytest.fixture(autouse=True)
def change_root(tmp_path: Path, monkeypatch: pytest.MonkeyPatch) -> None:
    """Automatically change directory to temp test directory."""
    monkeypatch.chdir(tmp_path)


@pytest.fixture
def app_without_root() -> t.Iterable[RootlessApp]:
    """App object without root."""
    yield startup({})


@pytest.fixture
def app(app_without_root: App) -> t.Iterable[App]:
    """App object with root."""
    app = app_without_root
    app.args = {"quiet": True}
    init(app)
    yield app


@pytest.fixture
def files_abc(tmp_path: Path) -> t.Iterable[t.List[Path]]:
    """Create files in tmp_path: a.md, b.md, c.md."""
    files = [tmp_path/"a.md", tmp_path/"b.md", tmp_path/"c.md"]
    for path in files:
        path.write_text(path.name)
    yield files



"""Test commands.py."""

from pathlib import Path
import pytest

from slipbox import commands
from slipbox.app import App, RootlessApp, startup
from slipbox.build import build
from slipbox.config import Config
from slipbox.dependencies import check_requirements


def scan(app: App) -> None:
    """Run 'slipbox build --no-output'."""
    app.args["output"] = False
    build(app)


def test_show_info_missing_note(app: App) -> None:
    """show_info should print error message and exit with error code."""
    app.args = {"note_id": 0}

    with pytest.raises(SystemExit) as system_exit:
        commands.show_info(app)

    assert system_exit.value.code != 0
    assert "does not exist" in system_exit.value.args[0]


@pytest.mark.skipif(
    not check_requirements(startup({})),
    reason="missing requirements",
)
def test_show_info_in_stdout(
    app: App,
    capsys: pytest.CaptureFixture[str],
) -> None:
    """show_info should output note info in stdout."""
    Path("test.md").write_text("# 0 Test\n\nTest note.", encoding="utf-8")

    build(app)
    app.args = {"note_id": 0}
    commands.show_info(app)
    stdout, stderr = capsys.readouterr()
    assert stdout
    assert not stderr


def test_init_creates_config_file(app_without_root: RootlessApp) -> None:
    """init must create .slipbox/config.cfg."""
    app = app_without_root
    commands.init(app)
    assert app.root is not None

    hidden = app.root/".slipbox"
    assert hidden.is_dir()
    assert hidden.joinpath("config.cfg").is_file()


def test_init_quiet(
    app_without_root: App,
    capsys: pytest.CaptureFixture[str],
) -> None:
    """init must not print anything in quiet mode."""
    app = app_without_root
    app.args["quiet"] = True
    commands.init(app)
    stdout, stderr = capsys.readouterr()
    assert not stdout
    assert not stderr


def test_init_root(app_without_root: RootlessApp) -> None:
    """init must set app.root."""
    app = app_without_root
    assert app.root is None
    commands.init(app)
    assert app.root is not None


def test_init_patterns(app_without_root: RootlessApp) -> None:
    """.slipbox/config.cfg must contain some glob patterns."""
    app = app_without_root
    commands.init(app)
    assert app.root is not None
    hidden = app.root/".slipbox"

    config = Config.from_file(hidden/"config.cfg")
    assert config.patterns["*.md"]
    assert config.patterns["*.rst"]


def test_init_already_initialized(app: App) -> None:
    """init must exit with error."""
    with pytest.raises(SystemExit) as system_exit:
        commands.init(app)

    assert system_exit.value.code != 0
    assert "has already been initialized" in system_exit.value.args[0]


def test_init_invalid_config(app_without_root: RootlessApp) -> None:
    """slipbox must not crash, but must exit with error instead."""
    Path("slipbox.cfg").write_text("invalid config file", encoding="utf-8")

    app = app_without_root
    app.args["config"] = "slipbox.cfg"

    with pytest.raises(SystemExit) as system_exit:
        commands.init(app)

    assert system_exit.value.code != 0
    assert "invalid config file" in system_exit.value.args[0]


@pytest.mark.skipif(
    not check_requirements(startup({})),
    reason="missing requirements",
)
class TestsWithRequirements:
    """Tests with external requirements (e.g. pandoc, graphviz, etc.)."""
    def test_check_notes_with_enabled_checks(
        self,
        app: App,
        capsys: pytest.CaptureFixture[str],
    ) -> None:
        """Only the enabled checks should be used.

        Checks that are enabled by default but aren't specified shouldn't be
        used.
        """
        Path("test.md").write_text("""
# 0 Foo

[Foo]().

# 1 Bar

Bar.

""", encoding="utf-8")
        app.args["enable"] = "missing-citations"
        build(app)
        capsys.readouterr()
        commands.check_notes(app)

        stdout, stderr = capsys.readouterr()
        assert not stdout
        assert not stderr

    def test_check_notes_with_disabled_checks(
        self,
        app: App,
        capsys: pytest.CaptureFixture[str],
    ) -> None:
        """Disabled checks shouldn't be used."""
        Path("test.md").write_text("""
# 0 Foo

[Foo]().

[Bar](#1).

# 1 Bar

Bar.
""", encoding="utf-8")
        build(app)
        capsys.readouterr()

        stdout, stderr = capsys.readouterr()
        assert not stdout
        assert not stderr

        app.args["disable"] = "empty-link-target"
        app.error_formatter.reset()
        commands.check_notes(app)

        stdout, stderr = capsys.readouterr()
        assert not stdout   # empty-link-target should be ignored.
        assert not stderr

    def test_check_notes_with_enabled_and_disabled_checks(
        self,
        app: App,
        capsys: pytest.CaptureFixture[str],
    ) -> None:
        """Only enabled checks that aren't disabled should be used."""
        Path("test.md").write_text("""
# 0 Foo

Foo. []()

# 1 Bar

Bar.

# 2 Baz.
""", encoding="utf-8")
        build(app)
        capsys.readouterr()

        stdout, stderr = capsys.readouterr()
        assert not stdout
        assert not stderr

        # Only empty-link-target should be used.
        app.args["enable"] = "isolated-note,empty-link-target"
        app.args["disable"] = "isolated-note"
        app.error_formatter.reset()
        commands.check_notes(app)

        stdout, stderr = capsys.readouterr()
        assert stdout
        assert not stderr

        assert "Empty link target" in stdout
        assert "#0" in stdout
        assert "Foo" in stdout
        assert "test.md" in stdout

        assert "#1" not in stdout
        assert "Bar" not in stdout

        assert "#2" not in stdout
        assert "Baz" not in stdout

    def test_check_notes_strict(
        self,
        app: App,
        capsys: pytest.CaptureFixture[str],
    ) -> None:
        """strict=True should turn warnings into errors."""
        Path("test.md").write_text("# 0 Test.", encoding="utf-8")
        app.args["strict"] = True
        scan(app)

        # isolated-note is normally a warning, but should be an error if
        # strict flag is turned on.
        with pytest.raises(SystemExit) as system_exit:
            commands.check_notes(app)

        assert system_exit.value.code != 0

        stdout, stderr = capsys.readouterr()
        assert stdout
        assert not stderr

        assert "Found errors :(" in stdout
        assert "error" in stdout
        assert "warning" not in stdout



"""Utils."""

import contextlib
import os
from pathlib import Path
import shlex
import subprocess
import sys
import tempfile
import typing as t


@contextlib.contextmanager
def temporary_directory() -> t.Iterator[Path]:
    """Path to temporary directory."""
    with tempfile.TemporaryDirectory() as tempdir:
        yield Path(tempdir)


def run_command(cmd: str,
                variables: t.Optional[t.Dict[str, str]] = None,
                **kwargs: t.Any) -> int:
    """Run command with additional environment variables in variables.

    Output stdout and stderr, and return the error code.

    kwargs
    : Additional arguments to subprocess.run
    """
    env = os.environ.copy()
    if variables is not None:
        env.update(variables)
    proc = subprocess.run(
        shlex.split(cmd),
        env=env,
        check=False,
        capture_output=True,
        **kwargs,
    )
    if proc.stdout:
        print(proc.stdout.decode())
    if proc.stderr:
        print(proc.stderr.decode(), file=sys.stderr)
    return proc.returncode


def show_error(verbosity: t.Literal["error", "warning"], message: str) -> None:
    """Print error message to stderr."""
    print(f"[{verbosity}]", message, file=sys.stderr)



"""Process CSV data."""

import csv
from pathlib import Path
from sqlite3 import Connection, IntegrityError
import typing as t

from .utils import show_error


def run_sql_on_csv(conn: Connection,
                   path: Path,
                   sql: str,
                   types: t.Sequence[t.Type[t.Any]],
                   callback: t.Optional[t.Callable[..., t.Any]] = None
                   ) -> None:
    """Run SQl query on CSV data."""
    cur = conn.cursor()
    with open(path, encoding="utf-8") as file:
        reader = csv.reader(file)
        for row in reader:
            args = [t(a) for t, a in zip(types, row)]
            try:
                cur.execute(sql, args)
            except IntegrityError:
                if callback:
                    callback(*args)


def process_files(conn: Connection, path: Path) -> None:
    """Process Files data in path."""
    sql = "INSERT OR IGNORE INTO Files (filename, hash) VALUES (?, ?)"
    run_sql_on_csv(conn, path, sql, (str, str))


def process_notes(conn: Connection, path: Path) -> bool:
    """Process Notes data in path.

    Returns False on error.
    """
    is_ok = True

    def fix(nid: int, title: str, filename: str) -> None:
        nonlocal is_ok
        is_ok = False
        cur = conn.cursor()
        cur.execute("SELECT title, filename FROM Notes WHERE id = ?", (nid,))
        existing = cur.fetchone()
        message = f"""Duplicate note ID (#{nid}). See:
- {existing[0]} ({existing[1]})
- {title} ({filename})"""
        show_error("error", message)

    sql = "INSERT INTO Notes (id, title, filename) VALUES (?, ?, ?)"
    run_sql_on_csv(conn, path, sql, (int, str, str), fix)
    return is_ok


def process_tags(conn: Connection, path: Path) -> None:
    """Process Tags data in path."""
    sql = "INSERT OR IGNORE INTO Tags (tag, id) VALUES (?, ?)"
    run_sql_on_csv(conn, path, sql, (str, int))


def process_links(conn: Connection, path: Path) -> None:
    """Process Links data in path."""
    sql = "INSERT OR IGNORE INTO Links (src, dest, direction) VALUES (?, ?, ?)"
    run_sql_on_csv(conn, path, sql, (int, int, str))


def process_bibliography(conn: Connection, path: Path) -> None:
    """Process Bibliography data in path."""
    sql = "INSERT OR IGNORE INTO Bibliography (key, html) VALUES (?, ?)"
    run_sql_on_csv(conn, path, sql, (str, str))


def process_citations(conn: Connection, path: Path) -> None:
    """Process Citations data in path."""
    sql = "INSERT OR IGNORE INTO Citations (note, reference) VALUES (?, ?)"
    run_sql_on_csv(conn, path, sql, (int, str))


def process_images(conn: Connection, path: Path) -> None:
    """Process Images data in path."""
    sql = "INSERT OR IGNORE INTO Images (filename, binary) VALUES (?, ?)"
    cur = conn.cursor()
    basedir = path.parent
    with open(path, encoding="utf-8") as file:
        reader = csv.reader(file)
        for row in reader:
            filename = row[0]
            image = basedir/filename
            binary: bytes = b""
            try:
                binary = image.read_bytes()
            except FileNotFoundError:
                continue
            try:
                cur.execute(sql, (filename, binary))
            except IntegrityError:
                pass


def process_image_links(conn: Connection, path: Path) -> None:
    """Process ImageLinks data in path."""
    sql = "INSERT OR IGNORE INTO ImageLinks (note, image) VALUES (?, ?)"
    run_sql_on_csv(conn, path, sql, (int, str))


def process_csvs(conn: Connection, basedir: Path) -> bool:
    """Process CSV data in basedir.

    Returns False on error.
    """
    process_files(conn, basedir/"files.csv")
    if not process_notes(conn, basedir/"notes.csv"):
        return False
    process_tags(conn, basedir/"tags.csv")
    process_links(conn, basedir/"links.csv")
    process_images(conn, basedir/"images.csv")
    process_image_links(conn, basedir/"image_links.csv")
    process_bibliography(conn, basedir/"bibliography.csv")
    process_citations(conn, basedir/"citations.csv")
    return True



"""Test config.py."""

from pathlib import Path

from slipbox.config import Config


def test_config_paths(tmp_path: Path) -> None:
    """Paths shouldn't be resolved in config file."""
    file = tmp_path/"config.cfg"

    config = Config()
    config.output_directory = Path("public")
    config.write(file)

    result = Config.from_file(file)
    output_directory = Path(result.output_directory)
    assert str(output_directory) == "public"


def test_config_read() -> None:
    """note-patterns should be read correctly.

    Patterns that are not in the config file should be implicitly ignored.
    """
    Path("config.cfg").write_text("""
[note-patterns]
*.txt = True
notes/*.tex = True
""", encoding="utf-8")

    config = Config.from_file(Path("config.cfg"))

    assert not config.patterns.get("*.md")
    assert config.patterns.get("*.txt")
    assert config.patterns.get("notes/*.tex")


def test_config_from_file_note_patterns() -> None:
    """Config.from_file shouldn't overwrite defaults if empty."""
    path = Path("config.cfg")
    path.touch()
    config = Config.from_file(path)

    default = Config()
    assert config == default



"""Test errors.py."""

from pathlib import Path

from slipbox.errors import ErrorFormatter, resolve_checkers


def test_error_formatter_duplicate_messages() -> None:
    """Duplicate messages should be minimized."""
    Path("test.json").write_text("""
[
    {
        "name": "duplicate-note-id",
        "value": {
            "id": 0,
            "notes": [
                {
                    "title": "Foo",
                    "filename": "foo.md"
                },
                {
                    "title": "Bar",
                    "filename": "bar.md"
                }
            ]
        }
    },
    {
        "name": "duplicate-note-id",
        "value": {
            "id": 0,
            "notes": [
                {
                    "title": "Foo",
                    "filename": "foo.md"
                },
                {
                    "title": "Baz",
                    "filename": "baz.md"
                }
            ]
        }
    }
]
""", encoding="utf-8")
    formatter = ErrorFormatter()
    formatter.add_errors(Path("test.json"))

    result = formatter.format().strip()
    assert "#0" in result
    assert "Foo" in result
    assert "Bar" in result
    assert "Baz" in result
    assert "foo.md" in result
    assert "bar.md" in result
    assert "baz.md" in result
    assert "Found errors :(" in result


def test_error_formatter_warnings_only() -> None:
    """'Found errors :(' text shouldn't appear in output."""
    Path("test.json").write_text("""
[
    {
        "name": "empty-link-target",
        "value": {
            "id": 0,
            "title": "Test",
            "filename": "test.md"
        }
    }
]
""", encoding="utf-8")
    formatter = ErrorFormatter()
    formatter.add_errors(Path("test.json"))

    result = formatter.format().strip()
    assert "Found errors" not in result


def test_resolve_checkers_default() -> None:
    """When args are None, should return default checkers."""
    checkers = resolve_checkers(None, None)
    assert checkers
    assert "" not in checkers


def test_resolve_checkers_enabled_only() -> None:
    """Result should be set of checkers in comma-separated list."""
    enabled = "graph-cycle,isolated-note"
    checkers = resolve_checkers(enabled, None)
    assert set(enabled.split(",")) == checkers
    assert "" not in checkers


def test_resolve_checkers_disabled_only() -> None:
    """Result should contain default checkers - disabled checkers."""
    disabled = "graph-cycle,isolated-note"
    checkers = resolve_checkers(None, disabled)
    assert checkers
    assert "" not in checkers
    assert "graph-cycle" not in checkers
    assert "isolated-note" not in checkers


def test_resolve_checkers_default_all() -> None:
    """Default checkers are a subset of all checkers."""
    default = resolve_checkers(None, None)
    all_ = resolve_checkers("all", None)
    assert default.issubset(all_)



"""Test utils.py."""
# pylint: disable=consider-using-f-string

from pathlib import Path
import shlex
import shutil

import pytest

from slipbox import utils


def test_temporary_directory() -> None:
    """temporary_directory gets deleted when the context manager exits."""
    with utils.temporary_directory() as temp:
        path = temp
        assert temp.exists()
        assert temp.is_dir()
    assert not path.exists()


@pytest.mark.skipif(not shutil.which("grep"), reason="requires grep")
def test_run_command(tmp_path: Path,
                     capsys: pytest.CaptureFixture[str],
                     ) -> None:
    """run_command must return returncode and capture stdout and stderr."""
    first = tmp_path/"first.txt"
    second = tmp_path/"second.txt"
    first.write_text("first")
    second.write_text("second")

    retcode = utils.run_command("grep first {} {}".format(
        shlex.quote(str(first)), shlex.quote(str(second))
    ))

    stdout, stderr = capsys.readouterr()
    assert not stderr
    assert not retcode
    assert str(first) in stdout
    assert str(second) not in stdout

    retcode = utils.run_command("grep second {} {}".format(
        shlex.quote(str(first)), shlex.quote(str(second))
    ))
    stdout, stderr = capsys.readouterr()
    assert not stderr
    assert str(first) not in stdout
    assert str(second) in stdout


@pytest.mark.skipif(not shutil.which("env"), reason="requires env")
def test_run_command_with_kwargs(capsys: pytest.CaptureFixture[str]) -> None:
    """Keyword arguments to run_command must be used as environment variables.
    """
    retcode = utils.run_command("env", dict(ZZZZZZZZZZ="ZZZZZZZZZZ"))
    stdout, stderr = capsys.readouterr()
    assert not stderr
    assert not retcode
    assert "ZZZZZZZZZZ=ZZZZZZZZZZ" in stdout



"""Test templates.py."""

from slipbox.templates import Elem, render


def test_render_text() -> None:
    """String must render as itself."""
    text = "text"
    assert render(text) == text


def test_render_element() -> None:
    """Element must render with children and attributes."""
    elem = Elem("a", "link", href="#")
    assert render(elem) == """<a href='#'>
  link
</a>"""


def test_render_childless_element() -> None:
    """Resulting string must be a single-line."""
    elem = Elem("div", **{"class": "button-group"})
    assert render(elem) == "<div class='button-group'></div>"


def test_render_nested() -> None:
    """Nested element must render with proper indentation."""
    elem = Elem("p", Elem("a", "link", href="#"))
    assert render(elem) == """<p>
  <a href='#'>
    link
  </a>
</p>"""


def test_render_no_attributes() -> None:
    """Element with no attributes must have no space after tag."""
    elem = Elem("p", "text")
    assert render(elem) == """<p>
  text
</p>"""



"""App object."""

import configparser
from dataclasses import dataclass, field
from functools import wraps
from pathlib import Path
from sqlite3 import Connection, connect
import sys
import typing as t

from .config import Config
from .database import migrate
from .errors import ErrorFormatter


def is_root(path: Path) -> bool:
    """Check if path is the root of a slipbox directory.

    NOTE Does not check if contents of .slipbox/ are valid.
    """
    return (path/".slipbox").is_dir()


def find_root() -> t.Optional[Path]:
    """Find slipbox root if it exists."""
    root = Path().resolve()
    while root != root.parent:
        if is_root(root):
            return root
        root = root.parent
    return root if is_root(root) else None


@dataclass
class RootlessApp:
    """App object before being initialized with a root directory."""
    args: t.Dict[str, t.Any]
    root: t.Optional[Path]
    config: Config
    database: Connection

    error_formatter: ErrorFormatter = field(default_factory=ErrorFormatter)

    def cleanup(self) -> None:
        """Clean up."""
        self.database.close()

    def backup_database(self) -> Path:
        """Create database backup.

        Returns path to backup database.
        """
        assert self.root
        backup = self.root/".slipbox"/"data.db.bak"
        with connect(str(backup.resolve())) as destination:
            self.database.backup(destination)
        return backup

    def restore_database_backup(self) -> None:
        """Restore database backup."""
        assert self.root
        self.cleanup()
        database = self.root/".slipbox"/"data.db"
        backup = self.root/".slipbox"/"data.db.bak"
        backup.replace(database)
        self.database = connect(str(database.resolve()))


@dataclass
class App(RootlessApp):
    """App object with a root directory."""
    root: Path


# Show error message and exit.
error = sys.exit


def startup(args: t.Dict[str, t.Any]) -> t.Union[App, RootlessApp]:
    """Prepare app."""
    app = RootlessApp(
        args=args,
        root=find_root(),
        config=Config(),
        database=connect(":memory:"),
    )

    root = app.root
    if root:
        try:
            app.config = Config.from_file(root/".slipbox"/"config.cfg")
        except configparser.Error:
            error("invalid config file: .slipbox/config.cfg")
        app.database = connect(root/".slipbox"/"data.db")
        app = t.cast(App, app)

    migrate(app.database)
    return app


Command = t.Callable[[RootlessApp], None]


def require_init(cmd: t.Callable[[App], None]) -> Command:
    """Wrap command that needs to be initialized."""
    @wraps(cmd)
    def wrapper(app: RootlessApp) -> None:
        if app.root is None:
            error("slipbox has not been initialized")
        return cmd(t.cast(App, app))
    return wrapper



"""Generate note graphs and graph layouts."""

import json
from sqlite3 import Connection
import typing as t

import networkx as nx   # type: ignore
from pyquery import PyQuery     # type: ignore

from .serializer import serialize


def create_graph(con: Connection) -> nx.DiGraph:
    """Construct graph from slipbox data."""
    graph = nx.DiGraph()

    sql = "SELECT id, filename FROM Notes"
    for id_, filename in con.execute(sql):
        graph.add_node(id_, filename=filename, tags=[])

    sql = "SELECT tag, id FROM Tags"
    for tag, id_ in con.execute(sql):
        graph.nodes[id_]["tags"].append(tag)

    sql = "SELECT src, dest, direction FROM ValidLinks"
    for src, dest, direction in con.execute(sql):
        if direction == "<":
            graph.add_edge(dest, src)
        else:
            graph.add_edge(src, dest)
    return graph


def find_cycles(graph: nx.DiGraph) -> t.List[t.Tuple[int, int]]:
    """Return a cycle in the graph, if there's any.

    The result is a list of links (pairs of ints).
    """
    try:
        return t.cast(t.List[t.Tuple[int, int]], nx.find_cycle(graph))
    except nx.exception.NetworkXNoCycle:
        return []


def concatenate(seqs: t.Iterable[t.Iterable[t.Any]]) -> t.List[t.Any]:
    """Concatenate sequences."""
    return sum(map(list, seqs), [])


def get_cluster(graph: nx.DiGraph, tag: str) -> nx.DiGraph:
    """Get subgraph with tag plus neighbors (preds and succs)."""
    tagged = [
        n for n, attrs in graph.nodes.items()
        if tag in attrs.get("tags")
    ]
    predecessors = concatenate(map(graph.predecessors, tagged))
    successors = concatenate(map(graph.successors, tagged))
    return graph.subgraph(tagged + predecessors + successors)


def get_components(graph: nx.DiGraph) -> t.Dict[t.Tuple[int, ...], nx.DiGraph]:
    """Get notes connected to note_id."""
    components = nx.algorithms.weakly_connected_components(graph)
    return {tuple(c): graph.subgraph(c) for c in components}


def without_self_loop(graph: nx.DiGraph) -> nx.DiGraph:
    """Return copy of graph without self-loops."""
    copy = graph.copy()
    copy.remove_edges_from(e for e in graph.edges if e[0] == e[1])
    return copy


def get_note_titles(con: Connection) -> t.Iterable[t.Tuple[int, str]]:
    """Get note title HTMLs."""
    sql = "SELECT id, html FROM Notes"
    for id_, html in con.execute(sql):
        doc = PyQuery(html)
        title = doc("h1:first").outer_html()
        yield (id_, title)


def compute_graph_layout(
    graph: nx.DiGraph,
    layout: str = "fdp",
) -> t.Dict[int, t.Tuple[float, float]]:
    """Compute graph layout using graphviz without using cache.

    Expects graph without self-loops.
    """
    positions = nx.drawing.nx_pydot.graphviz_layout(graph, prog=layout)
    return t.cast(t.Dict[int, t.Tuple[float, float]], positions)


def get_cached_graph_layout(
    con: Connection,
    key: str,
) -> t.Optional[t.Dict[int, t.Tuple[float, float]]]:
    """Return cached graph layout in json, or None."""
    sql = "SELECT layout FROM LayoutCache WHERE key = ?"

    cur = con.cursor()
    cur.execute(sql, (key,))
    row = cur.fetchone()
    if not row:
        return None

    cached = json.loads(row[0])
    return {int(k): v for k, v in cached.items()}


def save_graph_layout(con: Connection, key: str, value: str) -> None:
    """Save key value pair in LayoutCache.

    Assume that value is a valid JSON.
    Note: JSON keys are always strings.

    Caller is expected to commit changes afterward.
    """
    sql = """
        INSERT INTO LayoutCache (key, layout) VALUES (?, ?)
        ON CONFLICT (key) DO
        UPDATE SET layout = excluded.layout
        """
    con.execute(sql, (key, value))


def get_graph_layout(
    con: Connection,
    graph: nx.DiGraph,
    layout: str = "fdp",
) -> t.Dict[int, t.Tuple[float, float]]:
    """Get graph layout from LayoutCache or compute using graphviz.

    Expects graph without self-loops.
    """
    serialized = serialize(graph)
    cached = get_cached_graph_layout(con, serialized)
    if cached is not None:
        return cached

    result = compute_graph_layout(graph, layout)
    save_graph_layout(con, serialized, json.dumps(result))
    return result


def create_graph_data(
    con: Connection,
    titles: t.Dict[int, str],
    graph: nx.DiGraph,
    layout: str = "fdp",
) -> t.Dict[str, t.Any]:
    """Create cytoscape.js graph data from Graph.

    Note: removes self-loops.
    Prepares the graph layout and unpads title HTML.

    titles: dict of note IDs and title HTML fragments
    """
    copy = without_self_loop(graph)
    data = nx.readwrite.json_graph.cytoscape_data(copy)
    positions = get_graph_layout(con, copy, layout)

    for node in data["elements"]["nodes"]:
        node_id = int(node["data"]["id"])
        node["data"]["title"] = titles[node_id]
        x, y = positions[node_id]   # pylint: disable=invalid-name
        node["position"] = {"x": 2 * x, "y": -2 * y}
        # Scale y by -2 to flip the graph vertically so edges point downward.
    return t.cast(t.Dict[str, t.Any], data)



"""Get available note ID."""

import itertools
import typing as t

from ..app import App, require_init


def missing_integers(integers: t.Sequence[int]) -> t.Iterator[int]:
    """Yield integers that are not in the input.

    Assume the integers in the input are sorted.
    """
    if integers:
        yield from range(1, integers[0])
        for i in range(1, len(integers)):
            assert integers[i-1] <= integers[i]
            yield from range(integers[i-1] + 1, integers[i])
    yield from itertools.count(1 if not integers else integers[-1] + 1)


def take(count: int, items: t.Iterable[t.Any]) -> t.Iterator[t.Any]:
    """Take 'count' items from 'items'."""
    return itertools.islice(items, count)


@require_init
def new_note(app: App) -> None:
    """Get unused note IDs."""
    count = app.args["n"]
    rows = app.database.execute("SELECT id FROM Notes ORDER BY (id)")
    ids = [row[0] for row in rows]
    suggest = take(max(1, count), missing_integers(ids))
    print(",".join(map(str, suggest)))



"""Functions for interacting with sqlite3 database."""

from pathlib import Path
import re
from sqlite3 import Connection
import typing as t


def user_version(con: Connection) -> int:
    """Return database user_version."""
    cur = con.cursor()
    cur.execute("PRAGMA user_version")
    row = cur.fetchone()
    return 0 if not row else int(row[0])


class Schema(t.NamedTuple):
    """Schema object."""
    version: int
    source: str


def schemas() -> t.List[Schema]:
    """Return list of schemas sorted by version.

    Does not check if there are duplicate versions or if the file versions
    match the user version.
    """
    migrations = Path(__file__).with_name("migrations")
    pattern = re.compile(r"^(\d+)\..+\.sql$")
    scripts = []
    for path in migrations.glob("*.sql"):
        result = pattern.match(path.name)
        if result:
            scripts.append(Schema(int(result.group(1)), path.read_text()))

    scripts.sort()
    return scripts


def migrate(con: Connection) -> None:
    """Migrate to latest version of database.

    Does not roll back on error.
    """
    for version, source in schemas():
        if user_version(con) < version:
            con.executescript(source)
    con.commit()


__all__ = ["migrate"]



# pylint: disable=missing-docstring,too-many-branches,too-many-locals
"""Collects and formats error messages."""

import json
from pathlib import Path
import typing as t

import colorama  # type: ignore
from colorama import Fore, Style


colorama.init()


class Note(t.TypedDict):
    id: int
    title: str
    filename: str


class DuplicateNoteIDValue(t.TypedDict):
    id: int
    notes: t.List[Note]


class DuplicateNoteIDSchema(t.TypedDict):
    name: t.Literal["duplicate-note-id"]
    value: DuplicateNoteIDValue


class EmptyTargetLinkSchema(t.TypedDict):
    name: t.Literal["empty-link-target"]
    value: Note


class GraphCycleSchema(t.TypedDict):
    name: t.Literal["graph-cycle"]
    value: Note


class InvalidLinkValue(t.TypedDict):
    note: Note
    target: int


class InvalidLinkSchema(t.TypedDict):
    """Note links to non-existent note."""
    name: t.Literal["invalid-link"]
    value: InvalidLinkValue


class IsolatedNoteSchema(t.TypedDict):
    """Note is not reachable from other notes."""
    name: t.Literal["isolated-note"]
    value: Note


class MissingCitationsSchema(t.TypedDict):
    name: t.Literal["missing-citations"]
    value: Note


MessageSchema = t.Union[
    DuplicateNoteIDSchema,
    EmptyTargetLinkSchema,
    GraphCycleSchema,
    InvalidLinkSchema,
    IsolatedNoteSchema,
    MissingCitationsSchema,
]


def red(text: str) -> str:
    """Color text red."""
    return t.cast(str, Fore.RED + Style.BRIGHT + text + Style.RESET_ALL)


def yellow(text: str) -> str:
    """Color text yellow."""
    return t.cast(str, Fore.YELLOW + Style.BRIGHT + text + Style.RESET_ALL)


def bright(text: str) -> str:
    """Brighten text."""
    return t.cast(str, Style.BRIGHT + text + Style.RESET_ALL)


def dim(text: str) -> str:
    """Dim text."""
    return t.cast(str, Style.DIM + text + Style.RESET_ALL)


def is_error(message: MessageSchema) -> bool:
    """Check if message describes an error."""
    return message["name"] in (
        "duplicate-note-id",
        "invalid-link",
    )


def format_line(note: Note, info: str = "") -> str:
    """Format line in ErrorFormatter output.

    info: Optional info text to include.
    """
    id_ = dim(f"#{note['id']}")
    title = note["title"]
    filename = dim(f"({note['filename']})")

    result = f"  {id_} {title} {filename}"
    if info:
        result += f"\n    {info}"
    return result


def format_section(
    notes: t.Iterable[Note],
    header: str,
    footer: str = "",
    info: t.Optional[t.Dict[int, str]] = None,
) -> str:
    """Format section in ErrorFormatter output.

    If there are no notes, returns an empty string.
    """
    if not notes:
        return ""

    if info is None:
        info = {}

    section = header.strip() + "\n\n"
    written = set()
    for note in notes:
        id_ = note["id"]
        line = format_line(note, info=info.get(id_, "")) + "\n"
        if line in written:
            continue
        written.add(line)

        section += line

    section += "\n"
    if footer:
        section += f"  {footer.strip()}\n\n"
    return section


def resolve_checkers(
    enabled: t.Optional[str],
    disabled: t.Optional[str],
) -> t.Set[str]:
    """Determine set of checkers to use."""
    default = {
        "empty-link-target",
        "isolated-note",
        "missing-citations",
    }
    if enabled is None and disabled is None:
        return default

    excluded = set(disabled.split(",")) if disabled else set()
    if enabled is None:
        return default - excluded

    included = set(enabled.split(",") if enabled else set())
    checkers = included - excluded
    if "all" not in checkers:
        return checkers
    return {
        "empty-link-target",
        "graph-cycle",
        "isolated-note",
        "missing-citations",
    }


class ErrorFormatter:
    """Collects and formats error messages."""
    def __init__(self) -> None:
        self.messages: t.List[MessageSchema] = []

    def reset(self) -> None:
        """Clear messages."""
        self.messages = []

    def format(
        self,
        enabled: t.Optional[str] = None,
        disabled: t.Optional[str] = None,
        strict: bool = False,
    ) -> str:
        """Minimized output for all errors and warnings."""
        checkers = resolve_checkers(enabled, disabled)
        notes: t.Dict[str, t.List[Note]] = {
            "duplicate-note-id": [],
            "empty-link-target": [],
            "graph-cycle": [],
            "invalid-link": [],
            "isolated-note": [],
            "missing-citations": [],
        }

        invalid_link_targets: t.Dict[int, t.List[int]] = {}

        for message in self.messages:
            name = message["name"]
            value = message["value"]

            if name == "duplicate-note-id":
                value = t.cast(DuplicateNoteIDValue, value)
                for note in value["notes"]:
                    notes[name].append(dict(
                        id=value["id"],
                        title=note["title"],
                        filename=note["filename"],
                    ))
            elif name == "invalid-link":
                value = t.cast(InvalidLinkValue, value)
                note = value["note"].copy()
                notes[name].append(note)
                invalid_link_targets.setdefault(int(note["id"]), []).append(
                    value["target"],
                )
            else:
                notes[name].append(t.cast(Note, value).copy())

        invalid_link_info = {}
        for id_, targets in invalid_link_targets.items():
            invalid_link_info[id_] = "-> " + " ".join(
                bright(f"#{target}") for target in targets
            )

        error = red("error")
        warning = yellow("warning")
        if strict:
            warning = error

        result = (
            format_section(
                notes["duplicate-note-id"],
                header=f"{error}: Duplicate note ID",
            ) +
            format_section(
                notes["invalid-link"],
                header=f"{error}: Invalid link",
                footer="These notes link to non-existent notes.",
                info=invalid_link_info,
            )
        )
        has_errors = bool(result)
        has_warnings = False

        if "empty-link-target" in checkers:
            has_warnings = True
            result += format_section(
                notes["empty-link-target"],
                header=f"{warning}: Empty link target",
                footer="These notes contain links with an empty target.",
            )
        if "graph-cycle" in checkers:
            has_warnings = True
            result += format_section(
                notes["graph-cycle"],
                header=f"{warning}: Graph cycle",
                footer="A cycle was found in the notes graph.",
            )
        if "isolated-note" in checkers:
            has_warnings = True
            result += format_section(
                notes["isolated-note"],
                header=f"{warning}: Isolated note",
                footer="These notes are not reachable from other notes.",
            )
        if "missing-citations" in checkers:
            has_warnings = True
            result += format_section(
                notes["missing-citations"],
                header=f"{warning}: Missing citations",
                footer="These notes do not cite sources.",
            )

        if has_errors or strict and has_warnings:
            result += "Found errors :(\n"
        return result

    def add_error(self, message: MessageSchema) -> bool:
        """Add warning/error message.

        Returns True if it's an error.
        """
        self.messages.append(message)
        return is_error(message)

    def add_errors(self, path: Path) -> bool:
        """Collect errors from json file in path.

        Returns True if errors are found.
        """
        messages = json.loads(path.read_text(encoding="utf-8"))
        self.messages.extend(messages)
        return any(is_error(m) for m in messages)



"""Graph serialiation."""

from itertools import chain
import typing as t

import networkx as nx   # type: ignore


def serialize(graph: nx.DiGraph) -> str:
    """Serialize directed graph."""
    return "\n".join(
        " ".join(chain([str(node)], map(str, sorted(graph.adj[node]))))
        for node in sorted(graph.nodes)
    )


def deserialize(adjacency: str) -> t.Optional[nx.DiGraph]:
    """Deserialize directed graph from adjacency list.

    Returns None on error.
    """
    graph = nx.DiGraph()
    for line in adjacency.splitlines():
        try:
            node, *dests = [int(node) for node in line.split()]
        except ValueError:
            return None
        graph.add_node(node)
        for dest in dests:
            graph.add_edge(node, dest)
    return graph


def are_equal_graphs(first: nx.DiGraph, second: nx.DiGraph) -> bool:
    """Do both graphs have the same set of nodes and the same set of edges?"""
    return bool(first.nodes == second.nodes) and \
        bool(first.edges == second.edges)



"""Test processor.py."""

from hashlib import sha256
from pathlib import Path
import typing as t

import pytest

from slipbox.app import App
from slipbox.processor import build_command, preprocess, MARKDOWN_TEMPLATE


def test_preprocess_markdown_with_sources(files_abc: t.List[Path],
                                          tmp_path: Path,
                                          ) -> None:
    """There must be a code block between each file section in the
    result.
    """
    _, *sources = files_abc
    sources[0].write_text("B")
    sources[1].write_text("C")

    content = preprocess(MARKDOWN_TEMPLATE, *sources, basedir=tmp_path)
    template = """
```
[slipbox-metadata]
filename={}
hash={}
```
"""

    assert template.format(
        str(sources[0].relative_to(tmp_path)),
        sha256(b"B").hexdigest()
    ) in content
    assert template.format(
        str(sources[1].relative_to(tmp_path)),
        sha256(b"C").hexdigest()
    ) in content


def test_preprocess_markdown_with_no_sources(tmp_path: Path) -> None:
    """It must return an empty string."""
    content = preprocess(MARKDOWN_TEMPLATE, basedir=tmp_path)
    assert not content


def test_build_command(app: App) -> None:
    """Sanity check for build_command."""
    input_file = app.root/"input.md"
    output = "output.html"
    input_file.touch()

    app.config.strip_comments = False

    cmd = build_command(app, input_file, output)
    assert str(input_file) in cmd
    assert f"-o {output}" in cmd

    assert "--strip-comments" not in cmd

    app.config.strip_comments = True
    cmd = build_command(app, input_file, output)
    assert "--strip-comments" in cmd


@pytest.mark.xfail
def test_build_command_when_input_file_does_not_exist(app: App) -> None:
    """build_command must fail if input file does not exist."""
    input_file = app.root/"input.md"
    build_command(app, input_file, "output.html")



"""Static site generator for Zettelkasten notes."""

import sys
import typing as t

from . import __version__, commands
from .app import error, RootlessApp, startup
from .build import build
from .cli import parse_args
from .dependencies import has_dot, has_pandoc
from .tools.new import new_note


Command = t.Callable[[RootlessApp], None]


handlers: t.Dict[str, Command] = {
    "build": build,
    "check": commands.check_notes,
    "formats": commands.list_supported_formats,
    "info": commands.show_info,
    "init": commands.init,
    "new": new_note,
}


def main() -> None:
    """Entrypoint."""
    app = startup(parse_args())
    if not has_pandoc(app):
        error("pandoc not found")
    if not has_dot(app):
        error("dot (graphviz) not found")

    if app.args.get("version"):
        print(f"""slipbox {__version__}
Copyright (C) 2020 Levi Gruspe
https://github.com/lggruspe/slipbox""")
        sys.exit()

    command = app.args["(command)"]
    if command is not None:
        handler = handlers[command]
        if handler is not None:
            handler(app)
    app.cleanup()


if __name__ == "__main__":
    main()



"""Test tools/new.py."""

import typing as t

from hypothesis import given, strategies as st
from slipbox.tools.new import missing_integers, take


@given(st.lists(st.integers(min_value=0, max_value=1000)).map(sorted))  # type: ignore # noqa
def test_missing_integers_are_not_in_the_input(integers: t.List[int]) -> None:
    """It should generate integers not in the input."""
    count = integers[-1] if integers else 10
    result = take(count, missing_integers(integers))
    assert all(x not in integers for x in result)


@given(st.lists(st.integers(min_value=0)).map(sorted))  # type: ignore # noqa
def test_missing_integers_are_increasing(integers: t.List[int]) -> None:
    """Generated numbers should be sorted in increasing order."""
    result = list(take(len(integers), missing_integers(integers)))
    for i in range(1, len(result)):
        assert result[i-1] < result[i]


@given(st.lists(st.integers(min_value=0)).map(sorted))  # type: ignore # noqa
def test_missing_integers_do_not_contain_zero(integers: t.List[int]) -> None:
    """Generated numbers shouldn't contain 0."""
    assert next(missing_integers(integers)) > 0



"""Site builder."""

from hashlib import sha256
from pathlib import Path
import typing as t

from .app import App, error, require_init
from .batch import group_by_file_extension
from .generator import compile_site
from .processor import process_batch


def find_notes(app: App) -> t.Iterable[Path]:
    """Find notes in slipbox directory."""
    assert app.root and app.root.is_dir()

    include: t.List[str] = []
    exclude: t.List[str] = []
    for pattern, true in app.config.patterns.items():
        (include if true else exclude).append(pattern)

    for path in app.root.rglob("*"):
        if not path.is_file():
            continue
        if not any(path.match(pat) for pat in include):
            continue
        if any(path.match(pat) for pat in exclude):
            continue
        yield path


def find_outdated_notes(app: App, notes: t.Iterable[Path]) -> t.Iterable[Path]:
    """Outdated notes: notes in database whose hash have changed.

    NOTE Does not check if the contents of the files changed.
    """
    digests = {p: sha256(p.read_bytes()).hexdigest() for p in notes}
    outdated = []
    sql = "SELECT filename, hash FROM Files"
    for filename, _hash in app.database.execute(sql):
        path = app.root/filename
        if digests.get(path) != _hash:
            outdated.append(filename)
    return outdated


def find_new_notes(app: App, notes: t.Iterable[Path]) -> t.Iterable[Path]:
    """Find notes that are not yet in the database."""
    assert app.root is not None
    sql = "SELECT filename FROM Files"
    in_db = set(r[0] for r in app.database.execute(sql))
    for path in notes:
        filename = str(path.relative_to(app.root))
        if filename not in in_db:
            yield path


def process_notes(app: App, notes: t.Iterable[Path]) -> bool:
    """Process new notes (save into database).

    Returns False on error.
    """
    is_ok = True
    new = find_new_notes(app, notes)
    for batch in group_by_file_extension(new):
        is_ok = is_ok and process_batch(app, batch)
    return is_ok


def delete_notes(app: App, notes: t.Iterable[Path]) -> None:
    """Delete notes from database."""
    cur = app.database.cursor()
    cur.execute("PRAGMA foreign_keys=ON")
    cur.executemany(
        "DELETE FROM Files WHERE filename IN (?)",
        ((filename,) for filename in notes),
    )
    app.database.commit()


@require_init
def build(app: App) -> None:
    """Build website."""
    backup = app.backup_database()
    notes = list(find_notes(app))
    outdated = find_outdated_notes(app, notes)
    delete_notes(app, outdated)
    is_ok = process_notes(app, notes)
    print(app.error_formatter.format(), end="")

    if not is_ok:
        app.restore_database_backup()
        error(1)
    compile_site(app)
    backup.unlink(missing_ok=True)


__all__ = ["build", "process_notes"]



"""Runtime dependency checks."""

from shutil import which

from .app import RootlessApp


def has_dot(app: RootlessApp) -> bool:
    """Check if dot is installed."""
    return bool(which(app.config.dot))


def has_pandoc(app: RootlessApp) -> bool:
    """Check if pandoc is installed."""
    return bool(which(app.config.pandoc))


def check_requirements(app: RootlessApp) -> bool:
    """Check if all runtime requirements are satisfied."""
    return has_pandoc(app) and has_dot(app)



"""Generate site in output directory."""

from contextlib import contextmanager
import json
from pathlib import Path
from shutil import copytree, move, rmtree
from sqlite3 import Connection
import typing as t

from .app import App
from .graph import (
    create_graph,
    create_graph_data,
    get_cluster,
    get_components,
    get_note_titles,
)
from .page import generate_index
from .utils import temporary_directory


data = Path(__file__).parent/"data"

Generator = t.Callable[[Path], None]


def clear(directory: Path) -> None:
    """Clear contents of directory.

    Create directory if it doesn't exist.
    """
    directory.mkdir(exist_ok=True)
    for child in directory.iterdir():
        if child.is_dir():
            rmtree(child, ignore_errors=True)
        else:
            child.unlink()


@contextmanager
def output_directory_proxy(path: Path) -> t.Iterator[Path]:
    """Create proxy for output directory.

    Afterwards path is emptied and all contents of tempdir are moved into path.
    """
    assert not path.exists() or path.is_dir()
    with temporary_directory() as tempdir:
        yield tempdir
        clear(path)
        for child in tempdir.iterdir():
            move(str(child), str(path))


class IndexGenerator:
    """Generates index.html."""
    def __init__(self, app: App):
        self.app = app

    def run(self, out: Path) -> None:
        """Generate index.html inside output directory."""
        generate_index(self.app, out)


def copy(source: Path, dest: Path) -> None:
    """Copy text from source to dest Path."""
    dest.write_bytes(source.read_bytes())


def generate_js(out: Path) -> None:
    """Copy slipbox.js inside output directory."""
    copy(data/"slipbox.js", out/"slipbox.js")
    copy(data/"slipbox.js.map", out/"slipbox.js.map")


def generate_css(out: Path) -> None:
    """Copy slipbox.css into output directory."""
    copy(data/"slipbox.css", out/"slipbox.css")
    copy(data/"slipbox.css.map", out/"slipbox.css.map")


def generate_favicons(out: Path) -> None:
    """Copy favicons."""
    favicons = Path(__file__).parent/"favicons"
    for path in favicons.iterdir():
        copy(path, out/path.name)


def copy_boxicons(out: Path) -> None:
    """Copy boxicons svgs."""
    copytree(data/"svg", out/"assets"/"boxicons"/"svg")


def copy_mathjax(out: Path) -> None:
    """Copy mathjax files."""
    copytree(data/"es5", out/"es5")


class ImagesGenerator:
    """Copies images from database into output directory."""
    def __init__(self, con: Connection):
        self.con = con

    def run(self, out: Path) -> None:
        """Copy images into output directory."""
        (out/"images").mkdir()
        sql = "SELECT filename, binary FROM Images"
        for filename, binary in self.con.execute(sql):
            image = out/filename
            image.write_bytes(binary)


class CytoscapeDataGenerator:
    """Generate JSONs for cytoscape.js."""
    def __init__(self, con: Connection):
        self.con = con
        self.graph = create_graph(con)

        self.titles = dict(get_note_titles(self.con))

    def write(self, path: Path, graph: t.Any, layout: str = "fdp") -> None:
        """Write graph JSON data to path."""
        graph_data = create_graph_data(self.con, self.titles, graph, layout)
        path.write_text(json.dumps(graph_data))

    def run(self, out: Path) -> None:
        """Generate JSONs for cytoscape.js in out/graph."""
        (out/"graph").mkdir()
        layout = "dot" if self.graph.order() < 100 else "fdp"
        self.write(out/"graph"/"data.json", self.graph, layout)

        (out/"graph"/"tag").mkdir()
        for tag, in self.con.execute("SELECT DISTINCT tag FROM Tags"):
            path = out/"graph"/"tag"/f"{tag[1:]}.json"
            self.write(path, get_cluster(self.graph, tag), "dot")

        (out/"graph"/"note").mkdir()
        for component, subgraph in get_components(self.graph).items():
            graph_data = create_graph_data(
                self.con,
                self.titles,
                subgraph,
                "dot",
            )
            for note_id in component:
                path = out/"graph"/"note"/f"{note_id}.json"
                path.write_text(json.dumps(graph_data))
        self.con.commit()


def compile_site(app: App) -> None:
    """Copy files into output directory."""
    assert app.root is not None
    con = app.database
    output_directory = app.root/app.config.output_directory

    with output_directory_proxy(output_directory) as tempdir:
        CytoscapeDataGenerator(con).run(tempdir)
        ImagesGenerator(con).run(tempdir)
        IndexGenerator(app).run(tempdir)
        generate_css(tempdir)
        generate_js(tempdir)
        generate_favicons(tempdir)
        copy_boxicons(tempdir)
        copy_mathjax(tempdir)



# pylint: disable=missing-function-docstring
"""Test serializer.py."""

from random import shuffle
import typing as t

from hypothesis import given, strategies as st
import networkx as nx   # type: ignore
import pytest

from slipbox.serializer import are_equal_graphs, serialize, deserialize


def random_nodes() -> st.SearchStrategy[t.List[int]]:
    """Strategy for generating list of ints with no duplicates."""
    return st.lists(st.integers(min_value=0), unique=True)


def random_edges() -> st.SearchStrategy[t.List[t.Tuple[int, int]]]:
    """Strategy for generating list of (int, int)s with no duplicates."""
    pairs = st.tuples(st.integers(min_value=0), st.integers(min_value=0))
    return st.lists(pairs, unique=True)


@st.composite
def random_graph(draw: st.DrawFn) -> nx.DiGraph:
    """Create random graph strategy for testing."""
    graph = nx.DiGraph()
    graph.add_nodes_from(draw(random_nodes()))
    graph.add_edges_from(draw(random_edges()))
    return graph


@pytest.mark.xfail
@given(random_graph())
def test_random_graph_produces_trivial_graphs(
    random_graph: nx.DiGraph,
) -> None:
    assert random_graph.edges


@pytest.mark.xfail
@given(random_graph())
def test_random_graph_produces_non_trivial_graphs(
    random_graph: nx.DiGraph,
) -> None:
    assert not random_graph.edges


@given(random_graph())
def test_serialize_and_deserialize_are_inverse_functions(
    random_graph: nx.DiGraph,
) -> None:
    serialized = serialize(random_graph)
    deserialized = deserialize(serialized)

    assert are_equal_graphs(random_graph, deserialized)
    assert serialized == serialize(deserialized)


@given(random_graph())
def test_equal_graphs_serialize_the_same(random_graph: nx.DiGraph) -> None:
    """E.g. even if the edges are shuffled."""
    nodes = list(random_graph.nodes)
    edges = list(random_graph.edges)
    shuffle(nodes)
    shuffle(edges)

    shuffled_graph = nx.DiGraph()
    shuffled_graph.add_nodes_from(nodes)
    shuffled_graph.add_edges_from(edges)

    assert serialize(random_graph) == serialize(shuffled_graph)


def assert_is_sorted(items: t.Sequence[int]) -> None:
    """Check if all items are in sorted order."""
    for i in range(1, len(items)):
        assert items[i-1] <= items[i]


@given(random_graph())
def test_serialized_elements_are_sorted(random_graph: nx.DiGraph) -> None:
    nodes = []
    lines = serialize(random_graph).splitlines()
    for line in lines:
        node, *rest = line.split()
        nodes.append(int(node))
        assert_is_sorted([int(node) for node in rest])
    assert_is_sorted(nodes)



"""Test __main__.py."""

import shutil
import sys

import pytest

from slipbox.__main__ import main


@pytest.mark.skipif(not shutil.which("pandoc"), reason="requires pandoc")
def test_main_no_args(
    capsys: pytest.CaptureFixture[str],
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    """main shouldn't exit with error."""
    monkeypatch.setattr(sys, "argv", [])

    main()
    stdout, stderr = capsys.readouterr()
    assert not stdout
    assert not stderr



"""Process notes and store results in database.

Preprocessors insert metadata blocks into notes as code blocks.
They are mainly used to pass note metadata to Pandoc filters even when
the input is the concatenation of several files.
"""

from hashlib import sha256
from pathlib import Path
import shlex
from sqlite3 import Connection
import sys
import typing as t

from lxml.html import HtmlElement  # type: ignore
from pyquery import PyQuery  # type: ignore

from . import utils
from .app import App
from .batch import Batch
from .data import process_csvs


DOKUWIKI_TEMPLATE = """
<code>
[slipbox-metadata]
{}
</code>
"""

# NOTE Some brackets are escaped.
LATEX_TEMPLATE = r"""
\begin{{verbatim}}
[slipbox-metadata]
{}
\end{{verbatim}}
"""

MARKDOWN_TEMPLATE = """
```
[slipbox-metadata]
{}
```
"""

MEDIAWIKI_TEMPLATE = """
<pre>[slipbox-metadata]
{}</pre>
"""

ORG_TEMPLATE = """
#+begin_example
[slipbox-metadata]
{}
#+end_example
"""

RST_TEMPLATE = """
.. code::

    [slipbox-metadata]
    {}

"""

T2T_TEMPLATE = MARKDOWN_TEMPLATE

TEXTILE_TEMPLATE = """
bc. [slipbox-metadata]
{}

"""

METADATA_TEMPLATES = {
    ".dokuwiki": DOKUWIKI_TEMPLATE,
    ".latex": LATEX_TEMPLATE,
    ".markdown": MARKDOWN_TEMPLATE,
    ".md": MARKDOWN_TEMPLATE,
    ".mdown": MARKDOWN_TEMPLATE,
    ".org": ORG_TEMPLATE,
    ".rst": RST_TEMPLATE,
    ".t2t": T2T_TEMPLATE,
    ".tex": LATEX_TEMPLATE,
    ".textile": TEXTILE_TEMPLATE,
    ".wiki": MEDIAWIKI_TEMPLATE,
}


def render_metadata(template: str, prefix: str = "", **fields: t.Any) -> str:
    """Render metadata code block using template."""
    body = ('\n' + prefix).join(
        f"{k}={v}"
        for k, v in fields.items()
    )
    return template.format(body)


def preprocess(template: str,
               *sources: Path,
               basedir: Path,
               prefix: str = "",
               ) -> str:
    """Preprocess notes (sources) by inserting code blocks generated
    using the template.
    """
    def preprocess_single(source: Path) -> str:
        """Preprocess a single file."""
        content = source.read_bytes()
        filename = str(source.relative_to(basedir))
        _hash = sha256(content).hexdigest()
        metadata = render_metadata(template, prefix, filename=filename,
                                   hash=_hash)
        return metadata + content.decode(encoding="utf-8")
    return "".join(preprocess_single(source) for source in sources)


def store_html(conn: Connection,
               html: str,
               sources: t.Sequence[Path]) -> None:
    """Insert HTML sections into Notes table."""
    if not html.strip() or not sources:
        return
    cur = conn.cursor()
    sql = "UPDATE Notes SET html = ? WHERE id = ?"

    def callback(_: int, elem: HtmlElement) -> None:
        id_ = elem.get("id", "")
        if id_.isdigit():
            cur.execute(sql, (PyQuery(elem).outer_html(), int(id_)))

    doc = PyQuery(html)
    sections = doc("section")
    sections.map(callback)


def create_preprocessed_input(
    tempdir: Path,
    batch: Batch,
    basedir: Path
) -> Path:
    """Create preprocessed input to be passed to Pandoc."""
    template = METADATA_TEMPLATES.get(batch.extension, MARKDOWN_TEMPLATE)
    prefix = "    " if batch.extension == ".rst" else ""
    path = tempdir/("input" + batch.extension)
    path.write_text(
        preprocess(template, *batch.paths, basedir=basedir, prefix=prefix),
        encoding="utf-8"
    )
    return path


def build_options(app: App) -> str:
    """Build list of options to pass to pandoc based on user config."""
    options = ""
    config = app.config

    if config.strip_comments:
        options += " --strip-comments "
    if config.bibliography is not None:
        path = shlex.quote(str((app.root/config.bibliography).resolve()))
        options += f" --bibliography {path} --citeproc "
    return options


def build_command(app: App, input_: Path, output: str) -> str:
    """Construct pandoc command to run on input."""
    basedir = app.root
    options = build_options(app)

    assert input_.exists()
    assert basedir

    data = Path(__file__).parent/"data"
    lua_filter = shlex.quote(str((data/"filter.lua").resolve()))
    cmd = f"{app.config.pandoc} {options} -L{lua_filter} --section-divs " \
        f" -Mlink-citations:true --mathjax " \
        "--resource-path {} -o {} --extract-media=''".format(
            shlex.quote(str(basedir.resolve())),
            shlex.quote(output),
        )
    return cmd + ' ' + shlex.quote(str(input_.resolve()))


def process_batch(app: App, batch: Batch) -> bool:
    """Process batch of input notes.

    Returns False on error.
    """
    assert app.root is not None
    with utils.temporary_directory() as tempdir:
        preprocessed = create_preprocessed_input(tempdir, batch, app.root)
        html = tempdir/"temp.html"
        cmd = build_command(app, preprocessed, str(html))
        retcode = utils.run_command(cmd, cwd=tempdir)
        if retcode:
            print("Scan failed.", file=sys.stderr)
            return False

        if app.error_formatter.add_errors(tempdir/"messages.json"):
            return False

        if not process_csvs(app.database, tempdir):
            return False
        store_html(app.database, html.read_text(encoding="utf-8"), batch.paths)
        app.database.commit()
    return True



# pylint: disable=too-many-public-methods
"""Test slipbox.py."""

from hashlib import sha256
from pathlib import Path
from sqlite3 import Connection
import typing as t

import pytest
from slipbox.app import App, startup
from slipbox.build import (
    build, delete_notes, find_notes, find_new_notes, find_outdated_notes,
    process_notes,
)
from slipbox.dependencies import check_requirements


def scan(app: App) -> None:
    """Run 'slipbox build --no-output'."""
    app.args["output"] = False
    build(app)


def insert_files(con: Connection, *files: Path, basedir: Path) -> None:
    """Insert files into database."""
    sql = "INSERT INTO Files (filename, hash) VALUES (?, ?)"
    con.executemany(sql, (
        (
            str(p.relative_to(basedir)),
            sha256(p.read_bytes()).hexdigest()
        )
        for p in files
    ))


def is_quiet(capsys: pytest.CaptureFixture[str]) -> bool:
    """Check if there's no output in stdout or stderr."""
    stdout, stderr = capsys.readouterr()
    return not stdout and not stderr


def test_find_new_notes(app: App) -> None:
    """find_new_notes must only return existing files that aren't yet in the
    database and match the input patterns (*.md by default).
    """
    present = app.root/"present.md"
    absent = app.root/"absent.md"
    directory = app.root/"directory"
    txt = app.root/"ignore.txt"
    present.touch()
    absent.touch()
    directory.mkdir()
    txt.touch()

    insert_files(app.database, present, basedir=app.root)

    new_notes = find_new_notes(app, find_notes(app))
    assert list(new_notes) == [absent]


def test_added_notes_pattern(app: App) -> None:
    """Results of find_new_notes must match the input pattern."""
    app.config.patterns = {"*.md": True, "*.txt": True}
    directory = app.root/"directory"
    markdown = app.root/"input.md"
    txt = app.root/"input.txt"
    tex = app.root/"ignore.tex"
    directory.mkdir()
    markdown.touch()
    txt.touch()
    tex.touch()

    new_notes = find_new_notes(app, find_notes(app))
    assert sorted(new_notes) == [markdown, txt]


def test_added_notes_in_db(app: App) -> None:
    """Results of find_new_notes must not already be in the database."""
    new = app.root/"new.md"
    skip = app.root/"skip.md"
    new.touch()
    skip.touch()
    insert_files(app.database, skip, basedir=app.root)

    new_notes = find_new_notes(app, find_notes(app))
    assert list(new_notes) == [new]


def test_added_notes_recursive(app: App) -> None:
    """find_new_notes must find files recursively."""
    directory = app.root/"directory"
    new = app.root/"new.md"
    directory.mkdir()
    new.touch()

    new_notes = find_new_notes(app, find_notes(app))
    assert list(new_notes) == [new]


def test_modified_notes(app: App) -> None:
    """find_new_notes must find modified notes after they are deleted from the
    database.
    """
    modified = app.root/"modified.md"
    not_modified = app.root/"not_modified.md"
    added = app.root/"added.md"

    modified.write_text("hello")
    not_modified.write_text("hello")

    insert_files(app.database, modified, not_modified, basedir=app.root)

    not_modified.write_text("hello")    # not modified
    added.write_text("new note")
    modified.write_text("modified")

    delete_notes(app, find_outdated_notes(app, find_notes(app)))

    new_notes = find_new_notes(app, find_notes(app))
    assert sorted(new_notes) == [added, modified]


def test_purge(app: App, files_abc: t.List[Path]) -> None:
    """Input files must be purged from the database."""
    a_md, b_md, c_md = files_abc
    insert_files(app.database, a_md, b_md, c_md, basedir=app.root)
    a_md.write_text(a_md.read_text())
    b_md.unlink()
    c_md.write_text("edit")
    delete_notes(app, find_outdated_notes(app, find_notes(app)))

    result = app.database.execute("SELECT filename FROM Files")
    remaining = sorted(filename for filename, in result)
    assert len(remaining) == 1


@pytest.mark.skipif(
    not check_requirements(startup({})),
    reason="missing requirements",
)
class TestsWithRequirements:
    """Tests with external requirements (e.g. pandoc, graphviz, etc.)."""
    def test_build_backup_database(self, app: App) -> None:
        """Backup database must be deleted after build."""
        build(app)
        backup = app.root/".slipbox"/"data.db.bak"
        assert not backup.exists()

    def test_run(
        self,
        files_abc: t.List[Path],
        capsys: pytest.CaptureFixture[str],
        app: App,
    ) -> None:
        """There must be no suggestions when running for the first time."""
        file_a, file_b, file_c = files_abc
        file_a.write_text("# 0 A\n\nA.\n[B](#2 '/a').\n")
        file_b.write_text("# 1 B\n\nB.\n[C](#2).\n")
        file_c.write_text("# 2 C\n\nC.\n")

        build(app)
        assert is_quiet(capsys)

    def test_process(self, app: App) -> None:
        """Smoke test for slipbox.process."""
        input_file = app.root/"input.md"
        input_file.write_text("# 1 Test note\n\nHello, world!\n")
        assert not list(app.database.execute(
            "SELECT * FROM Notes WHERE html IS NOT NULL"
        ))

        process_notes(app, [input_file])
        result = list(app.database.execute(
            "SELECT * FROM Notes WHERE html IS NOT NULL"
        ))
        assert result

    def test_process_empty_file(self, app: App) -> None:
        """Empty files shouldn't have entries in the database."""
        empty = app.root/"empty.md"
        empty.touch()
        process_notes(app, [empty])
        assert not list(app.database.execute("SELECT * FROM Files"))
        assert not list(app.database.execute("SELECT * FROM Notes"))
        assert not list(app.database.execute("SELECT * FROM Tags"))
        assert not list(app.database.execute("SELECT * FROM Links"))
        assert not list(app.database.execute("SELECT * FROM Bibliography"))
        assert not list(app.database.execute("SELECT * FROM Citations"))

    def test_process_clusters_from_context(
        self,
        app: App,
    ) -> None:
        """Check if clusters are stored in db."""
        markdown = app.root/"test.md"
        markdown.write_text("""# 0 Test

#test

[](#1).

# 1 Dest

Test.
""")
        process_notes(app, [markdown])
        result = sorted(app.database.execute("SELECT src, dest FROM Links"))
        assert result == sorted([(0, 1)])

        result = sorted(app.database.execute("SELECT tag, id FROM Tags"))
        assert result == sorted([("#test", 0)])

    def test_process_filenames(self, app: App) -> None:
        """Filenames must be scanned correctly."""
        markdown = app.root/"foo.md"
        skip = app.root/"bar.md"
        markdown.write_text("# 0 Note\n\nBody.\n")
        skip.write_text("# 0 Note\n\nBody.\n")
        process_notes(app, [markdown])

        sql = "SELECT filename FROM Notes WHERE id = 0"
        result = list(app.database.execute(sql))
        assert len(result) == 1
        assert markdown.samefile(app.root/result[0][0])

    def test_process_filenames0(self, app: App) -> None:
        """Filenames must be scanned correctly."""
        markdown = app.root/"bar.md"
        skip = app.root/"foo.md"
        markdown.write_text("# 0 Note\n\nBody.\n")
        skip.write_text("# 0 Note\n\nBody.\n")
        process_notes(app, [markdown])

        sql = "SELECT filename FROM Notes WHERE id = 0"
        result = list(app.database.execute(sql))
        assert len(result) == 1
        assert markdown.samefile(app.root/result[0][0])

    def test_process_non_level1_headers(self, app: App) -> None:
        """Only level 1 headers must be considered as note headers."""
        markdown = app.root/"test.md"
        markdown.write_text("""# 0 Valid note header

Foo.

## 1 Invalid note header

Bar.
""")
        process_notes(app, [markdown])
        result = [nid for nid, in app.database.execute("SELECT id FROM Notes")]
        assert len(result) == 1
        assert result == [0]

    def test_process_with_duplicate_existing_id(
        self,
        app: App,
        capsys: pytest.CaptureFixture[str],
    ) -> None:
        """slipbox.process must show a warning if a new note shares the ID of
        an existing note.

        The warning message must show the filenames of both notes.
        """
        file_a = app.root/"a.md"
        file_a.write_text("# 0 Existing note\n\nTest.\n")

        process_notes(app, [file_a])
        result = list(app.database.execute("SELECT id, title FROM Notes"))
        assert len(result) == 1
        assert result == [(0, "Existing note")]

        assert is_quiet(capsys)

        file_b = app.root/"b.md"
        file_b.write_text("# 0 Duplicate note\n\nTest.\n")

        process_notes(app, [file_b])
        result = list(app.database.execute("SELECT id, title FROM Notes"))
        assert len(result) == 1
        assert result == [(0, "Existing note")]

        stdout, stderr = capsys.readouterr()
        assert not stdout
        assert stderr
        assert str(file_a.relative_to(app.root)) in stderr
        assert str(file_b.relative_to(app.root)) in stderr

    def test_build_duplicate_tags(
        self,
        capsys: pytest.CaptureFixture[str],
        app: App,
    ) -> None:
        """Database should only have one entry for note-tag pair."""
        Path("test.md").write_text("""
# 0 Test

Test.

#test #test
""", encoding="utf-8")
        scan(app)
        assert is_quiet(capsys)

        result = list(app.database.execute("SELECT id, tag FROM Tags"))
        assert result == [(0, "#test")]

    def test_build_with_negative_id(self, app: App) -> None:
        """Negative numbers should not be recognized as an ID."""
        Path("test.md").write_text("""# -1 Invalid note

Invalid.

# 0 Valid note

[Not a note link](#-1)
.""", encoding="utf-8")
        build(app)

        result = list(app.database.execute("SELECT id, title FROM Notes"))
        assert result == [(0, "Valid note")]

        query = "SELECT src, dest, direction FROM Links"
        result = list(app.database.execute(query))
        assert not result

    def test_build_with_duplicate_ids_in_a_file(
        self,
        app: App,
        capsys: pytest.CaptureFixture[str],
    ) -> None:
        """If there are duplicate IDs in a file, save neither note.

        slipbox.build must show an error message when this happens,
        containing the note ID, the filename and the note titles.
        """
        Path("test.md").write_text("""# 0 First note

Foo.

# 0 Duplicate

Bar.
""", encoding="utf-8")
        with pytest.raises(SystemExit) as system_exit:
            build(app)

        assert system_exit.value.code != 0

        result = list(app.database.execute("SELECT id, title FROM Notes"))
        assert not result

        stdout, stderr = capsys.readouterr()
        assert "#0" in stdout
        assert "First note" in stdout
        assert "Duplicate" in stdout
        assert "test.md" in stdout
        assert not stderr

    def test_build_with_duplicate_ids_in_batch(
        self,
        app: App,
        capsys: pytest.CaptureFixture[str],
    ) -> None:
        """If there are duplicate IDs in a batch, it should exit with error.

        It should also display an error message containing the filenames
        containing the duplicate IDs, and make sure that the database isn't
        modified.
        """
        Path("foo.md").write_text("# 0 Foo\n\nFoo.", encoding="utf-8")
        Path("bar.md").write_text("# 0 Bar\n\nBar.", encoding="utf-8")

        before = "\n".join(app.database.iterdump())

        with pytest.raises(SystemExit) as system_exit:
            build(app)

        after = "\n".join(app.database.iterdump())

        assert before == after
        assert system_exit.value.code != 0

        stdout, stderr = capsys.readouterr()
        assert "foo.md" in stdout
        assert "bar.md" in stdout
        assert "Foo" in stdout
        assert "Bar" in stdout
        assert "#0" in stdout
        assert not stderr

    def test_build_with_duplicate_ids_in_multiple_batches(
        self,
        app: App,
        capsys: pytest.CaptureFixture[str],
    ) -> None:
        """Similar to test_build_with_duplicate_ids_in_batch.

        This time, with different note formats.
        """
        Path("foo.rst").write_text("""
0 Foo
=====

Foo.
""", encoding="utf-8")
        Path("bar.md").write_text("# 0 Bar\n\nBar.", encoding="utf-8")

        before = "\n".join(app.database.iterdump())

        with pytest.raises(SystemExit) as system_exit:
            build(app)

        after = "\n".join(app.database.iterdump())

        assert before == after
        assert system_exit.value.code != 0

        stdout, stderr = capsys.readouterr()
        assert not stdout
        assert "foo.rst" in stderr
        assert "bar.md" in stderr
        assert "Foo" in stderr
        assert "Bar" in stderr
        assert "#0" in stderr

    def test_build_with_codeblock(
        self,
        app: App,
        capsys: pytest.CaptureFixture[str],
    ) -> None:
        """CodeBlocks shouldn't crash metadata parser in filters."""
        Path("test.md").write_text("""# 0 Hello

```
print("Hello :)")
```

# 1 Bye

Bye-bye.
""")
        build(app)
        assert is_quiet(capsys)

    def test_build_with_empty_link_target(
        self,
        app: App,
        capsys: pytest.CaptureFixture[str],
    ) -> None:
        """slipbox.build must show a warning if there is a link with an empty
        target.

        The link should be saved in the database, but the dest value should be
        negative.
        """
        Path("test.md").write_text("# 0 Foo\n\n[Empty]().\n", encoding="utf-8")
        build(app)

        query = "SELECT src, dest, direction FROM Links"
        result = list(app.database.execute(query))
        assert result == [(0, -1, "")]

        stdout, stderr = capsys.readouterr()
        assert "#0" in stdout
        assert "Foo" in stdout
        assert "test.md" in stdout
        assert not stderr

    def test_process_with_external_links(
        self,
        capsys: pytest.CaptureFixture[str],
        app: App,
    ) -> None:
        """External links shouldn't be saved in the db."""
        markdown = app.root/"test.md"
        markdown.write_text("# 0 Test\n\n[Example](https://example.com)")
        process_notes(app, [markdown])

        result = list(app.database.execute("SELECT id, title FROM Notes"))
        assert (0, "Test") in result

        result = list(app.database.execute("SELECT * FROM Links"))
        assert not result

        assert is_quiet(capsys)

    def test_process_with_id_in_scientific_form(
        self,
        capsys: pytest.CaptureFixture[str],
        app: App,
    ) -> None:
        """Headers with non-integer IDs should be ignored."""
        markdown = app.root/"test.md"
        markdown.write_text("# 1e1 Invalid note ID\n\nTest.\n")
        process_notes(app, [markdown])

        result = list(app.database.execute("SELECT * FROM Notes"))
        assert not result

        assert is_quiet(capsys)

    def test_process_with_non_text_titles(
        self,
        capsys: pytest.CaptureFixture[str],
        app: App,
    ) -> None:
        """Notes with non-text titles in the header must still be recognized.
        """
        file_a = app.root/"a.md"
        file_b = app.root/"b.md"
        file_c = app.root/"c.md"
        file_a.write_text("# 0 $1 + 1 = 2$\n\nTest.\n")
        file_b.write_text("# 1 [Note 0](#0)\n\nTest.\n")
        file_c.write_text("# 2 `print('Title')`\n\nTest.\n")

        process_notes(app, [file_a, file_b, file_c])

        sql = "SELECT id, title FROM Notes ORDER BY id"
        result = list(app.database.execute(sql))

        assert result == [
            (0, "1 + 1 = 2"),
            (1, "Note 0"),
            (2, "print('Title')"),
        ]

        assert is_quiet(capsys)

    def test_process_tags_with_trailing_punctuation(
        self,
        capsys: pytest.CaptureFixture[str],
        app: App,
    ) -> None:
        """If a hashtag has trailing invalid symbols, only the prefix must be
        saved.
        """
        markdown = app.root/"test.md"
        markdown.write_text("# 0 Test\n\n#tag.\n#tags.\n#0.\n")
        process_notes(app, [markdown])
        result = app.database.execute("SELECT tag FROM Tags ORDER BY tag")
        assert list(result) == [("#0",), ("#tag",), ("#tags",)]

        assert is_quiet(capsys)

    def test_process_rst(
        self,
        capsys: pytest.CaptureFixture[str],
        app: App,
    ) -> None:
        """Slipbox filters shouldn't raise metadata-related errors."""
        rst = app.root/"test.rst"
        rst.write_text("""
1 Foo
=====

Foo.

2 Bar
=====

Bar.""")

        process_notes(app, [rst])
        result = list(
            app.database.execute("SELECT id, title, filename FROM Notes")
        )

        assert result == [(1, "Foo", "test.rst"), (2, "Bar", "test.rst")]

        assert is_quiet(capsys)


@pytest.mark.skipif(
    not check_requirements(startup({})),
    reason="missing requirements",
)
class TestDirectionalLinks:
    """Test directional links (e.g. [foo](<#123))."""

    def write_test_md(self, app: App) -> None:
        """Write test note to test.md."""
        app.root.joinpath("test.md").write_text("""
# 0 Foo

[Bar](#1)

# 1 Bar

[Baz](>#2)

# 2 Baz

[Foo](<#0)
""")

    def test_process_with_directional_links(
        self,
        app: App,
        capsys: pytest.CaptureFixture[str],
    ) -> None:
        """Direction in prefix of link target should be removed."""
        self.write_test_md(app)
        build(app)
        assert is_quiet(capsys)

        sql = "SELECT html FROM Notes"
        notes = list(app.database.execute(sql))
        assert notes
        for html, in notes:
            assert "%3" not in html
            assert "<#" not in html
            assert ">#" not in html

    def test_process_direction_in_db(
        self,
        app: App,
        capsys: pytest.CaptureFixture[str],
    ) -> None:
        """Link "direction" should be stored in the database."""
        self.write_test_md(app)
        build(app)
        assert is_quiet(capsys)

        sql = "SELECT direction FROM Links WHERE src = ? AND dest = ?"
        cur = app.database.cursor()
        cur.execute(sql, (0, 1))
        assert cur.fetchone()[0] == ""

        cur.execute(sql, (1, 2))
        assert cur.fetchone()[0] == ">"

        cur.execute(sql, (2, 0))
        assert cur.fetchone()[0] == "<"



# type: ignore
"""Test batch.py."""

from slipbox.batch import group_by_file_extension


def test_group_by_file_extension() -> None:
    """group_by_file_extension should split by file type.

    Files with no extension should be considered their own type.
    """
    files = ["a.md", "b.md", ".md", "c.tex", ".tex", ""]
    batches = group_by_file_extension(files)
    groups = [batch.paths for batch in batches]
    assert len(groups) == 5
    assert ("a.md", "b.md") in groups
    assert (".md",) in groups
    assert ("c.tex",) in groups
    assert (".tex",) in groups
    assert ("",) in groups


def test_group_by_file_extension_on_the_same_type() -> None:
    """group_by_file_extension should group files with the same type together.
    """
    files = tuple(f"{c}.md" for c in "abcdefg")
    batches = group_by_file_extension(files)
    groups = [batch.paths for batch in batches]
    assert len(groups) == 1
    assert files in groups



"""Test cli.py."""

from slipbox.cli import parse_args


def test_parse_args_version() -> None:
    """parse_args should not require subcommand."""
    parse_args(["-v"])
    parse_args(["--version"])



# pylint: disable=use-a-generator
"""Check slipbox notes."""

import typing as t

from .app import App
from .errors import (
    EmptyTargetLinkSchema,
    GraphCycleSchema,
    InvalidLinkValue,
    InvalidLinkSchema,
    IsolatedNoteSchema,
    MissingCitationsSchema,
    Note,
)
from .graph import create_graph, find_cycles


def check_empty_links(app: App) -> t.List[Note]:
    """Check for notes with empty links.

    Results are reflected in app.error_formatter.
    Returns list of notes with empty links.
    """
    notes = []
    sql = """
        SELECT DISTINCT id, title, filename
        FROM Links JOIN Notes ON src = id
        WHERE dest < 0
        ORDER BY id
    """
    for nid, title, filename in app.database.execute(sql):
        value: Note = dict(
            id=int(nid),
            title=title,
            filename=filename,
        )
        message: EmptyTargetLinkSchema = dict(
            name="empty-link-target",
            value=value,
        )
        app.error_formatter.add_error(message)
        notes.append(value)
    return notes


def check_graph_cycles(app: App) -> t.List[Note]:
    """Check for cycles in the notes graph.

    Results are reflected in app.error_formatter.
    Returns list of notes in the cycle.
    """
    cycle = []
    edges = find_cycles(create_graph(app.database))

    query = "SELECT id, title, filename FROM Notes WHERE id = ?"

    cur = app.database.cursor()
    for src, _ in edges:
        cur.execute(query, (src,))
        row = cur.fetchone()

        value: Note = dict(
            id=int(row[0]),
            title=row[1],
            filename=row[2],
        )
        message: GraphCycleSchema = dict(
            name="graph-cycle",
            value=value,
        )
        app.error_formatter.add_error(message)
        cycle.append(value)
    return sorted(cycle, key=lambda note: note["id"])


def check_invalid_links(app: App) -> t.List[InvalidLinkValue]:
    """Check for invalid links.

    Results are reflected in app.error_formatter.
    Returns list of invalid links.
    """
    # NOTE Ignores negative (empty) links
    sql = """
        SELECT DISTINCT id, title, filename, dest
        FROM Links JOIN Notes ON src = id
        WHERE dest >= 0 AND dest NOT IN (
            SELECT id FROM Notes
        )
        ORDER BY id
    """
    invalid_links = []
    for nid, title, filename, dest in app.database.execute(sql):
        value: InvalidLinkValue = dict(
            note=dict(
                id=int(nid),
                title=title,
                filename=filename,
            ),
            target=int(dest),
        )
        message: InvalidLinkSchema = dict(
            name="invalid-link",
            value=value,
        )
        app.error_formatter.add_error(message)
        invalid_links.append(value)
    return invalid_links


def check_isolated_notes(app: App) -> t.List[Note]:
    """Check for isolated notes.

    Results are reflected in app.error_formatter.
    Returns list of isolated notes.
    """
    isolated_notes = []
    sql = """
        SELECT DISTINCT id, title, filename FROM Notes
        WHERE id NOT IN (
            SELECT src FROM ValidLinks UNION SELECT dest FROM ValidLinks
        )
    """
    for nid, title, filename in app.database.execute(sql):
        value: Note = dict(
            id=int(nid),
            title=title,
            filename=filename,
        )
        message: IsolatedNoteSchema = dict(
            name="isolated-note",
            value=value,
        )
        app.error_formatter.add_error(message)
        isolated_notes.append(value)
    return isolated_notes


def check_unsourced_notes(app: App) -> t.List[Note]:
    """Check for notes with missing citations.

    This checker is only used if there's a bib file.
    Results are reflected in app.error_formatter.
    Returns list of unsourced notes.
    """
    if app.config.bibliography is None:
        return []

    unsourced_notes = []
    sql = """
        SELECT DISTINCT id, title, filename FROM Notes
        WHERE id NOT IN (
            SELECT note FROM Citations
        )
    """
    for nid, title, filename in app.database.execute(sql):
        value: Note = dict(
            id=int(nid),
            title=title,
            filename=filename,
        )
        message: MissingCitationsSchema = dict(
            name="missing-citations",
            value=value,
        )
        app.error_formatter.add_error(message)
        unsourced_notes.append(value)
    return unsourced_notes


def check_notes(app: App) -> bool:
    """Check notes in slipbox.

    Returns False is errors are found.
    """
    has_error = check_invalid_links(app)
    has_warning = False

    checkers = [
        check_empty_links,
        check_graph_cycles,
        check_isolated_notes,
        check_unsourced_notes,
    ]
    has_warning = any([check(app) for check in checkers])
    strict = app.args.get("strict", False)
    if has_error or has_warning:
        enabled = app.args.get("enable")
        disabled = app.args.get("disable")
        print(
            app.error_formatter.format(enabled, disabled, strict=strict),
            end="",
        )
    return not has_error and (not strict or not has_warning)



"""Functions for batching input files by extension."""

from dataclasses import dataclass
from itertools import groupby
import os
from pathlib import Path
import typing as t


@dataclass
class Batch:
    """Batch of input files with the same file extension."""
    extension: str
    paths: t.Sequence[Path]


def group_by_file_extension(files: t.Iterable[Path]) -> t.Iterable[Batch]:
    """Batch input files by file extension.

    Files with no extension belong to their own batch.
    """
    def key(filename: t.Union[str, Path]) -> t.Tuple[str, str]:
        root, ext = os.path.splitext(filename)
        return (ext, "") if ext else ("", root)
    groups = groupby(sorted(files, key=key), key=key)
    for _key, paths in groups:
        yield Batch(_key[0] or _key[1], tuple(paths))



"""Test app.py."""

import pytest

from slipbox.app import App, find_root, startup


def test_find_root_in_current(app: App) -> None:
    """find_root must find .slipbox if it's in the current directory."""
    assert find_root() == app.root


def test_find_root_in_parent(
    app: App,
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    """find_root must find .slipbox if it's in a parent directory."""
    path = app.root/"foo"/"bar"/"baz"
    path.mkdir(parents=True)
    monkeypatch.chdir(path)

    assert not path.joinpath(".slipbox").exists()

    root = find_root()
    assert root is not None
    assert app.root == root
    assert root.joinpath(".slipbox").exists()


def test_find_root_none() -> None:
    """find_root should return None if slipbox has not been initialized."""
    assert find_root() is None


def test_app_database_backup_and_restore(app: App) -> None:
    """Backup database should be removed after restoring."""
    backup = app.root/".slipbox"/"data.db.bak"
    before = "\n".join(app.database.iterdump())

    assert not backup.exists()

    app.backup_database()
    app.database.execute("drop table Notes")

    assert backup.is_file()

    app.restore_database_backup()
    after = "\n".join(app.database.iterdump())

    assert not backup.exists()
    assert before == after


def test_startup_with_existing_directory(app: App) -> None:
    """startup should find the notes directory root."""
    alias = startup({})
    assert alias.root == app.root


def test_startup_with_invalid_config(app: App) -> None:
    """The program should exit with non-zero error."""
    config = app.root/".slipbox"/"config.cfg"
    config.write_text("invalid config file :)")

    with pytest.raises(SystemExit) as system_exit:
        startup({})

    assert system_exit.value.code != 0
    assert "invalid config file" in system_exit.value.args[0]



"""setup.py for slipbox."""

from pathlib import Path
import re
import typing as t

import setuptools   # type: ignore


def get_version() -> str:
    """Get version string from slipbox/__init__.py."""
    pattern = re.compile(r'__version__ = "(.+)"')
    init = Path(__file__).with_name("slipbox")/"__init__.py"
    result = pattern.search(init.read_text())

    assert result
    return result.groups()[0]


def get_install_requires() -> t.List[str]:
    """Get install requires from requirements/install.requirements.txt."""
    path = Path(__file__).with_name("requirements")/"install.requirements.txt"
    lines = path.read_text().splitlines()
    return [">=".join(line.split("==")) for line in lines]


if __name__ == "__main__":
    setuptools.setup(
        name="slipbox",
        version=get_version(),
        author="Levi Gruspe",
        author_email="mail.levig@gmail.com",
        description="Static site generator for Zettelkasten notes",
        long_description=Path("README.md").read_text(encoding="utf-8"),
        long_description_content_type="text/markdown",
        url="https://github.com/lggruspe/slipbox",
        packages=setuptools.find_packages(),
        package_data={
            "slipbox": [
                "data/*",
                "data/*/*",
                "data/*/*/*",
                "data/*/*/*/*",
                "data/*/*/*/*/*",
                "data/*/*/*/*/*/*",
                "favicons/*",
                "migrations/*.sql",
                "templates/*",
            ]
        },
        classifiers=[
            "Environment :: Console",
            "License :: OSI Approved :: MIT License",
            "Programming Language :: Python :: 3.8",
            "Programming Language :: Python :: 3.9",
            "Programming Language :: Python :: 3.10",
            "Programming Language :: Python :: 3.11",
        ],
        install_requires=get_install_requires(),
        python_requires=">=3.8",
        entry_points={
            "console_scripts": [
                "slipbox=slipbox.__main__:main"
            ]
        }
    )



# pylint: disable=too-many-instance-attributes
"""Slipbox configuration.

Config overriding strategy:
command-line args > environment variables > config file > default config.
"""

from configparser import ConfigParser
from dataclasses import dataclass
import os
from pathlib import Path


@dataclass
class Config:
    """Slipbox config class."""
    # [slipbox]
    output_directory = Path("public")
    title = "Slipbox"

    # [note-patterns]
    patterns = {
        "*.md": True,
        "*.markdown": True,
        "*.rst": True,
    }

    # [paths]
    pandoc = "pandoc"
    dot = "dot"

    # [pandoc-options]
    bibliography = None
    strip_comments = True

    # [check]
    checks = {
        "empty-link-target": True,
        "isolated-note": True,
        "missing-citations": True,
    }

    def __post_init__(self) -> None:
        self._update_from_env()

    @staticmethod
    def from_file(path: Path) -> "Config":
        """Return Config object from file."""
        parser = ConfigParser()
        parser.read_string(path.read_text())

        default = Config()

        # [slipbox]
        default.output_directory = Path(
            parser.get(
                "slipbox",
                "output_directory",
                fallback=default.output_directory,
            ),
        )
        default.title = parser.get("slipbox", "title", fallback=default.title)

        # [paths]
        default.pandoc = parser.get("paths", "pandoc", fallback=default.pandoc)
        default.dot = parser.get("paths", "dot", fallback=default.dot)

        # [note-patterns]
        if parser.has_section("note-patterns"):
            default.patterns = {}
            for key, _ in parser.items("note-patterns"):
                val = parser.getboolean("note-patterns", key, fallback=False)
                default.patterns[key] = val

        # [pandoc-options]
        bibliography = parser.get(
            "pandoc-options",
            "bibliography",
            fallback=None,
        )
        if bibliography:
            default.bibliography = Path(bibliography)

        default.strip_comments = parser.getboolean(
            "pandoc-options",
            "strip-comments",
            fallback=default.strip_comments,
        )

        # [check]
        if parser.has_section("check"):
            default.checks = {}
            for key, _ in parser.items("check"):
                val = parser.getboolean("check", key, fallback=False)
                default.checks[key] = val

        default._update_from_env()  # pylint: disable=protected-access
        return default

    def _update_from_env(self) -> None:
        """Update some config variables from environment variables."""
        self.pandoc = os.getenv("PANDOC", self.pandoc)
        self.dot = os.getenv("DOT", self.dot)

    def to_config_parser(self) -> ConfigParser:
        """Convert to ConfigParser object."""
        config = ConfigParser()

        sections = [
            "slipbox",
            "note-patterns",
            "paths",
            "pandoc-options",
            "check",
        ]
        for section in sections:
            if not config.has_section(section):
                config.add_section(section)

        config.set(
            "slipbox",
            "output_directory",
            str(self.output_directory),
        )
        config.set("slipbox", "title", self.title)

        for pattern, include in self.patterns.items():
            config.set(
                "note-patterns",
                pattern,
                "true" if include else "false",
            )

        config.set("paths", "pandoc", self.pandoc)
        config.set("paths", "dot", self.dot)

        if self.bibliography is not None:
            config.set(
                "pandoc-options",
                "bibliography",
                str(self.bibliography),
            )
        config.set(
            "pandoc-options",
            "strip-comments",
            "true" if self.strip_comments else "false",
        )

        for check, include in self.checks.items():
            config.set(
                "check",
                check,
                "true" if include else "false",
            )
        return config

    def write(self, path: Path) -> None:
        """Write config to file."""
        with open(path, "w", encoding="utf-8") as file:
            self.to_config_parser().write(file)


__all__ = ["Config"]



"""Slipbox CLI parser."""

from argparse import ArgumentParser, RawDescriptionHelpFormatter
from pathlib import Path
import typing as t


def parse_args(argv: t.Optional[t.Sequence[str]] = None) -> t.Dict[str, t.Any]:
    """Returns dict of command-line options and arguments.

    Name of subcommand can be accessed using the key: "(command)".
    """
    epilog = """
  build     Build website.
  check     Check for isolated notes and invalid links.
  formats   List supported input file formats.
  info      Show note info.
  init      Initialize notes directory.
  new       Get unused note IDs.
"""

    parser = ArgumentParser(
        prog="slipbox",
        description="Generate a static website from your notes.",
        formatter_class=RawDescriptionHelpFormatter,
        epilog=epilog,
    )
    parser.add_argument(
        "-v",
        "--version",
        action="store_true",
        help="show version number and exit",
    )

    # Add subcommands
    subparsers = parser.add_subparsers(
        title="commands",
        dest="(command)",
    )

    subparser = subparsers.add_parser("build", description="Build website.")
    subparser.add_argument(
        "--no-output",
        action="store_false",
        dest="output",
        help="update database only; do not generate site in output directory",
    )

    subparser = subparsers.add_parser(
        "check",
        description="Check for isolated notes and invalid links.",
    )
    subparser.add_argument(
        "-e",
        "--enable",
        dest="enable",
        help="comma-separated list of checkers to enable (overrides config)",
    )
    subparser.add_argument(
        "-d",
        "--disable",
        dest="disable",
        help="comma-separated list of checkers to disable (overrides config)",
    )
    subparser.add_argument(
        "--strict",
        action="store_true",
        dest="strict",
        help="turn warnings into errors",
    )

    subparser = subparsers.add_parser("info", description="Show note info.")
    subparser.add_argument("note_id", type=int)

    subparsers.add_parser(
        "formats",
        description="List supported input file formats.",
    )

    subparser = subparsers.add_parser(
        "init",
        description="Initialize notes directory.",
    )
    subparser.add_argument("config", nargs="?", type=Path, help="config file")
    subparser.add_argument("-q", "--quiet", action="store_true")

    subparser = subparsers.add_parser(
        "new",
        description="Get unused note IDs.",
    )
    subparser.add_argument(
        "-n",
        type=int,
        default=1,
        help="number of note IDs to generate",
    )
    return vars(parser.parse_args(argv))


__all__ = ["parse_args"]



"""Test tools/check.py."""

from pathlib import Path

import pytest

from slipbox import check
from slipbox.app import App, startup
from slipbox.build import build
from slipbox.dependencies import check_requirements


def scan(app: App) -> None:
    """Run 'slipbox build --no-output'."""
    app.args["output"] = False
    build(app)


@pytest.mark.skipif(
    not check_requirements(startup({})),
    reason="missing requirements",
)
class TestsWithRequirements:
    """Tests with external requirements (e.g. pandoc, graphviz, etc.)."""
    def test_check_empty_links(self, app: App) -> None:
        """check_empty_links must return only notes with empty links."""
        Path("test.md").write_text(
            "# 0 Foo\n\nFoo. []()\n\n# 1 Bar\n\nBar.\n",
            encoding="utf-8",
        )
        scan(app)

        result = list(check.check_empty_links(app))
        assert result == [
            dict(
                id=0,
                title="Foo",
                filename="test.md",
            ),
        ]

    def test_check_invalid_links(self, app: App) -> None:
        """check_invalid_links must return note and invalid ID."""
        Path("test.md").write_text(
            "# 0 Test\n[](#1)\n\n# 2 Test\n[](#0)\n",
            encoding="utf-8",
        )
        scan(app)

        result = list(check.check_invalid_links(app))
        assert result == [
            dict(
                note=dict(
                    id=0,
                    title="Test",
                    filename="test.md",
                ),
                target=1,
            ),
        ]

    def test_check_invalid_links_with_empty_links(self, app: App) -> None:
        """check_invalid_links shouldn't raise errors for empty links."""
        Path("test.md").write_text("# 0 Test []()", encoding="utf-8")
        scan(app)

        result = list(check.check_invalid_links(app))
        assert not result

    def test_check_isolated_notes(self, app: App) -> None:
        """check_isolated_notes must return untagged notes only."""
        Path("test.md").write_text("""# 0 Foo

#test

[](#1)

# 1 Bar

# 2 Baz
""", encoding="utf-8")
        scan(app)

        result = list(check.check_isolated_notes(app))
        assert result == [
            dict(
                id=2,
                title="Baz",
                filename="test.md",
            ),
        ]

    def test_check_unsourced_notes_empty_bibliography(self, app: App) -> None:
        """check_unsourced_notes must be empty if there is no bibliography."""
        Path("test.md").write_text("# 0 Test\n\nTest.\n", encoding="utf-8")
        scan(app)

        result = list(check.check_unsourced_notes(app))
        assert not result

    def test_check_unsourced_notes(self, app: App) -> None:
        """check_unsourced_notes must include every note without a citation."""
        app.config.bibliography = Path("test.bib")

        Path("test.bib").write_text("""
@book{test_2020,
    title = {Title},
    language = {English},
    author = {Author},
    year = {2020},
}
""")

        Path("test.md").write_text("""# 0 Foo

Foo.

# 1 Bar

Bar.

[@test_2020]
""", encoding="utf-8")
        scan(app)

        result = list(check.check_unsourced_notes(app))
        assert result == [
            dict(
                id=0,
                title="Foo",
                filename="test.md",
            ),
        ]

    def test_check_notes_empty(
        self,
        app: App,
        capsys: pytest.CaptureFixture[str],
    ) -> None:
        """check_notes must not output anything if there are no errors.

        The result must be True (no errors).
        """
        assert check.check_notes(app)
        stdout, stderr = capsys.readouterr()
        assert not stdout
        assert not stderr

    def test_check_notes(
        self,
        app: App,
        capsys: pytest.CaptureFixture[str],
    ) -> None:
        """check_notes must output to stdout.

        The result must be False (has errors).
        """
        Path("test.md").write_text("# 0 Test\n[](#1)", encoding="utf-8")
        scan(app)

        assert not check.check_notes(app)
        stdout, stderr = capsys.readouterr()
        assert stdout
        assert "Test" in stdout
        assert not stderr

    def test_check_notes_with_empty_links(
        self,
        app: App,
        capsys: pytest.CaptureFixture[str],
    ) -> None:
        """check_notes must include empty-link-target check."""
        Path("test.md").write_text("# 0 Test\n[]()", encoding="utf-8")
        scan(app)
        capsys.readouterr()

        stdout, stderr = capsys.readouterr()
        assert not stdout
        assert not stderr

        app.error_formatter.reset()
        assert check.check_notes(app)
        stdout, stderr = capsys.readouterr()

        assert "Empty link target" in stdout
        assert "#0" in stdout
        assert "Test" in stdout
        assert "test.md" in stdout

        assert not stderr

    def test_check_notes_graph_cycle_default(
        self,
        app: App,
        capsys: pytest.CaptureFixture[str],
    ) -> None:
        """graph-cycle check should be disabled by default."""
        Path("test.md").write_text("# 0 Test\n[](#0)", encoding="utf-8")
        scan(app)

        assert check.check_notes(app)
        stdout, stderr = capsys.readouterr()
        assert not stdout
        assert not stderr

    def test_check_notes_graph_cycle_enabled_with_one_note(
        self,
        app: App,
        capsys: pytest.CaptureFixture[str],
    ) -> None:
        """check_notes should catch graph cycles."""
        Path("test.md").write_text("# 0 Test\n[](#0)", encoding="utf-8")
        app.args["enable"] = "graph-cycle"
        scan(app)

        assert check.check_notes(app)
        stdout, stderr = capsys.readouterr()
        assert stdout
        assert not stderr

        assert "Graph cycle" in stdout
        assert "#0" in stdout
        assert "Test" in stdout
        assert "test.md" in stdout
        assert "warning" in stdout
        assert "error" not in stdout

    def test_check_notes_graph_cycle_enabled_with_multiple_notes(
        self,
        app: App,
        capsys: pytest.CaptureFixture[str],
    ) -> None:
        """Only notes in the cycle should appear in the warning message."""
        Path("test.md").write_text("""
# 0 Foo

Foo. [](#1)

# 1 Bar

Bar. [](#0)

# 2 Baz

Baz.
""", encoding="utf-8")
        app.args["enable"] = "graph-cycle"
        scan(app)

        assert check.check_notes(app)
        stdout, stderr = capsys.readouterr()
        assert stdout
        assert not stderr

        assert "Graph cycle" in stdout
        assert "test.md" in stdout

        assert "#0" in stdout
        assert "Foo" in stdout

        assert "#1" in stdout
        assert "Bar" in stdout

        assert "#2" not in stdout
        assert "Baz" not in stdout

        assert "warning" in stdout
        assert "error" not in stdout

    def test_check_notes_all_graph_cycle(
        self,
        app: App,
        capsys: pytest.CaptureFixture[str],
    ) -> None:
        """'all' should enable 'graph-cycle'."""
        Path("test.md").write_text("# 0 Test\n[](#0)", encoding="utf-8")
        app.args["enable"] = "all"
        scan(app)

        assert check.check_notes(app)
        stdout, stderr = capsys.readouterr()
        assert stdout
        assert not stderr

        assert "Graph cycle" in stdout
        assert "#0" in stdout
        assert "Test" in stdout
        assert "test.md" in stdout
        assert "warning" in stdout
        assert "error" not in stdout

    def test_check_notes_with_multiple_errors(
        self,
        app: App,
        capsys: pytest.CaptureFixture[str],
    ) -> None:
        """`slipbox check` should catch all errors, not just the first one."""
        Path("test.md").write_text("""
# 0 Foo

[]()

[](#0)
""", encoding="utf-8")
        app.args["enable"] = "all"
        scan(app)

        assert check.check_notes(app)
        stdout, stderr = capsys.readouterr()

        assert stdout
        assert not stderr

        assert "Empty link target" in stdout
        assert "Graph cycle" in stdout



"""Basic templates.

Also includes functions for building HTML elements.
"""

from pathlib import Path
from string import Template as StringTemplate
from textwrap import indent
import typing as t


class Template(StringTemplate):
    """Extended string.Template with file loader."""
    @staticmethod
    def load(path: Path, encoding: t.Optional[str] = None) -> "Template":
        """Load template from file."""
        return Template(path.read_text(encoding=encoding))

    def render(self, **kwargs: t.Any) -> str:
        """Substitute template variables."""
        return self.safe_substitute(**kwargs)


def render_template(path: t.Union[str, Path], **kwargs: t.Any) -> str:
    """Render template from templates directory."""
    basedir = Path(__file__).parent/"templates"
    template = Template.load(basedir/path, encoding="utf-8")
    return template.render(**kwargs)


class Elem:
    """Represents an HTML element."""
    def __init__(self,
                 tag: str,
                 *children: t.Union[str, "Elem"],
                 **attributes: t.Any):
        self.tag = tag
        self.children = list(children)
        self.attributes = attributes.copy()


def render(elem: t.Union[str, Elem], prefix: str = "") -> str:
    """Render HTML from Elem or str."""
    if isinstance(elem, str):
        return prefix + elem
    assert isinstance(elem, Elem)
    attrs = ""
    for key, val in elem.attributes.items():
        attrs += f" {key}={val!r}"
    if not elem.children:
        return "<{tag}{attrs}></{tag}>".format(tag=elem.tag, attrs=attrs)
    children = indent("\n".join(render(child) for child in elem.children),
                      prefix + "  ")
    return "<{tag}{attrs}>\n{children}\n</{tag}>".format(
        tag=elem.tag,
        attrs=attrs,
        children=children)



"""Generate complete HTML containing all slipbox notes."""
# pylint: disable=consider-using-f-string

from pathlib import Path
import shlex
from sqlite3 import Connection
import subprocess
import typing as t

from pyquery import PyQuery as pq  # type: ignore

from .app import App
from .templates import Elem, render, render_template
from .utils import temporary_directory


def render_dummy(title: str) -> str:
    """Render dummy markdown."""
    return fr"""---
title: {title}
...

::: {{style="display:none"}}
```c
```
:::
"""


class Note(t.NamedTuple):
    """Data needed to render list of notes."""
    id: int
    title_html: str


def note_list_item(note: Note) -> str:
    """Return slipbox-list item HTML for given note."""
    html = render_template("list__item.html", id=note.id, html=note.title_html)
    return html.rstrip()


def note_list(notes: t.Iterable[Note]) -> str:
    """Render slipbox-list HTML for given notes."""
    items = map(note_list_item, sorted(notes))
    return render_template("list.html", items="\n".join(items)).rstrip()


def get_section_title(html: str) -> str:
    """Get innerHTML of section title."""
    assert html is not None
    doc = pq(html)
    header = doc("h1")
    return t.cast(str, header.html())


def render_home(conn: Connection, title: str) -> str:
    """Create home page HTML section containing a list of all notes."""
    notes = (
        Note(id_, get_section_title(html))
        for id_, html in conn.execute("SELECT id, html FROM Notes")
    )
    list_ = note_list(notes)
    return render_template("home.html", title=title, list=list_)


def generate_active_htmls(conn: Connection) -> t.Iterable[str]:
    """Get HTML stored in the database for active sections."""
    sql = "SELECT html FROM Notes WHERE html IS NOT NULL ORDER BY id ASC"
    return (html.strip() for html, in conn.execute(sql))


def render_references(conn: Connection) -> str:
    """Create bibliography HTML section from database entries."""
    sql = "SELECT key, html FROM Bibliography ORDER BY key"
    items = '\n'.join(
        render_template("bibliography__item.html", **dict(
            href=f"#{key}", term=f"[@{key[4:]}]", description=html
        )).strip()
        for key, html in conn.execute(sql)
    )
    return render_template("bibliography.html", items=items)


def render_untagged(con: Connection) -> str:
    """Render HTML for list of untagged notes."""
    rows = con.execute("SELECT id, html FROM Untagged")
    notes = list(
        Note(id_, get_section_title(html))
        for id_, html in rows
    )
    if not notes:
        return ""
    return render_template("untagged.html", notes=note_list(notes))


def render_tags(conn: Connection) -> str:
    """Create HTML section that lists all tags.

    Also list untagged notes.

    Note: notes without a tag but are connected to a tagged note are considered
    "weakly tagged".
    """
    rows = conn.execute(
        "SELECT tag, COUNT(*) FROM Tags GROUP BY tag ORDER BY tag"
    )
    tags = (
        f'<li><a href="#tags/{tag[1:]}">{tag}</a> ({count})</li>'
        for tag, count in rows
    )
    return render_template(
        "tags.html",
        tags="\n".join(tags),
        untagged=render_untagged(conn),
    )


def render_tag_page(conn: Connection, tag: str) -> str:
    """Create HTML section that lists all notes with the tag."""
    sql = """
        SELECT id, html FROM Tags NATURAL JOIN Notes WHERE tag = ? ORDER BY id
    """
    notes = (
        Note(id_, get_section_title(html))
        for id_, html in conn.execute(sql, (tag,))
    )
    section = Elem("section",
                   Elem("h1", tag),
                   note_list(notes),
                   id=f"tags/{tag[1:]}",
                   title=tag,
                   **{"class": "level1"})
    return render(section)


def render_tag_pages(conn: Connection) -> str:
    """Create all tag pages."""
    rows = conn.execute("SELECT DISTINCT tag FROM Tags ORDER BY tag")
    tags = (row[0] for row in rows)
    return '\n'.join(render_tag_page(conn, tag) for tag in tags)


def render_reference_page(conn: Connection, reference: str) -> str:
    """Create HTML section that lists all notes that cite the reference."""
    sql = """
        SELECT note, Bibliography.html,Notes.html FROM Citations
            JOIN Notes ON Citations.note = Notes.id
                JOIN Bibliography ON Bibliography.key = Citations.reference
                    WHERE reference = ?
                        ORDER BY note
    """
    notes = []
    text = ""
    for note, _text, html in conn.execute(sql, (reference,)):
        assert not text or text == _text
        text = _text
        notes.append(Note(note, get_section_title(html)))
    section = Elem("section",
                   Elem("h1", '@' + reference[4:]),
                   Elem("p", text),
                   note_list(notes),
                   id=reference,
                   title=reference,
                   **{"class": "level1"})
    return render(section)


def render_reference_pages(conn: Connection) -> str:
    """Create all reference pages."""
    rows = conn.execute("SELECT key FROM Bibliography ORDER BY key")
    references = (row[0] for row in rows)
    return '\n'.join(render_reference_page(conn, ref) for ref in references)


def render_main(conn: Connection, title: str = "Slipbox") -> str:
    """Main content of index.html."""
    return render_template(
        "main.html",
        nav=render_template("nav.html"),
        home=render_home(conn, title),
        sections="\n".join(generate_active_htmls(conn)),
        tag_pages=render_tag_pages(conn),
        tags=render_tags(conn),
        reference_pages=render_reference_pages(conn),
        references=render_references(conn),
    )


def _write(path: Path, text: str) -> None:
    """Write text to file in path."""
    path.write_text(text, encoding="utf-8")


def generate_index(app: App, out: Path) -> None:
    """Create final HTML file with javascript."""
    con = app.database
    options = "-s"
    title = app.config.title

    with temporary_directory() as tempdir:
        _write(tempdir/"header.txt", render_template("header.html"))
        _write(tempdir/"Slipbox.md", render_dummy(title))
        _write(tempdir/"after.txt", render_main(con, title))

        cmd = """{pandoc} Slipbox.md -Hheader.txt --metadata title:{title}
                -Aafter.txt --section-divs {opts} -o {output} -c slipbox.css
            """.format(
            pandoc=app.config.pandoc,
            title=shlex.quote(title),
            opts=options,
            output=shlex.quote(str(out/"index.html")))
        subprocess.run(shlex.split(cmd), check=False, cwd=tempdir)
